"""Module with general utilities."""

import os as _os
import sys as _sys
import time as _time
from collections import Counter as _Counter
import logging as _log
import inspect as _inspect
import subprocess as _sp
import datetime as _datetime

import epics as _epics

from mathphys import beam_optics as _beam
from siriuspy import envars as _envars


def conv_splims_labels(label):
    """Convert setpoint limit labels from pcaspy to epics and vice-versa."""
    # Limits description:
    # -------------------
    # DRVH     driver high level limit
    # HIHI     high-high level limit (ALARM)
    # HIGH     IOC high level limit (ALARM)
    # DRVL     driver low level limit
    # LOLO     IOC low-low level limit (ALARM)
    # LOW      IOC low level limit (ALARM)
    # LOPR     Low operating range
    # HOPR     High operating range
    #
    # Rules
    # -----
    # LOLO <= LO <= HI <= HIHI
    # DRVL <= LOPR <= HOPR <= DRVH.
    #
    labels_dict = {
        # Epics-DB  pcaspy    PyEpics
        # ===========================
        'DRVH':     'DRVH',   # upper_ctrl_limit
        'HIHI':     'hihi',   # upper_alarm_limit
        'HIGH':     'high',   # upper_warning_limit
        'HOPR':     'hilim',  # upper_disp_limit
        'LOPR':     'lolim',  # lower_disp_limit
        'LOW':      'low',    # lower_warning_limit
        'LOLO':     'lolo',   # lower_alarm_limit
        'DRVL':     'DRVL',   # lower_ctrl_limit
        'TSTV':     'TSTV',   # SIRIUS specific (Test value)
        'TSTR':     'TSTR',   # SIRIUS specific (Test acceptable range)
    }
    if label in labels_dict:
        # epics -> pcaspy
        return labels_dict[label]
    else:
        for k, value in labels_dict.items():
            if value == label:
                # pcaspy -> epics
                return k
        raise KeyError('Invalid splims label "' + label + '"!')


def get_last_commit_hash():
    """Get commit Hash of the repository of the calling file."""
    fname = _os.path.realpath(_inspect.stack()[1][1])
    path = fname.rpartition(_os.path.sep)[0]
    pwd = _os.path.abspath('.')
    return _sp.getoutput('cd ' + path +
                         '; git log --format=%h -1; cd ' + pwd)


def get_timestamp(now=None):
    """Get formatted timestamp ."""
    if now is None:
        now = _time.time()
    rst = _datetime.datetime.fromtimestamp(now).strftime('%Y-%m-%d-%H:%M:%S')
    rst = rst + '.{0:03d}'.format(int(1000*(now-int(now))))
    return rst


def read_text_data(text):
    """Parse text data from sirius-consts web server."""
    lines = text.splitlines()
    parameters = {}
    data = []
    for line in lines:
        line = line.strip()
        if not line:
            continue  # empty line
        if line[0] == '#':
            if len(line[1:].strip()) > 0:
                token, *words = line[1:].split()
                if token[0] == '[':
                    # it is a parameter.
                    parm = token[1:-1].strip()
                    parameters[parm] = words
        else:
            # it is a data line
            data.append(line.split())
    return data, parameters


def print_ioc_banner(ioc_name, db, description, version, prefix):
    """IOC banner."""
    strl = '==================================='
    nchars = (len(strl)-len(ioc_name))//2
    line = ' '*nchars + ioc_name + ' '*nchars
    _log.info(strl)
    _log.info(line)
    _log.info(strl)
    _log.info(description)
    _log.info('LNLS, Sirius Project.')
    _log.info('Version   : %s', version)
    _log.info('Timestamp : %s', get_timestamp())
    _log.info('Prefix    : %s', prefix)
    _log.info('')
    pvs = sorted(tuple(db.keys()))
    for i, pvname in enumerate(pvs, 1):
        _log.info('{0:04d} {1:<}'.format(i, pvname))


def configure_log_file(stream=None, filename=None, debug=False):
    """Configure logging messages for the IOCs."""
    if stream is not None:
        dic_ = {'stream': stream}
    elif filename is not None:
        dic_ = {'filename': filename, 'filemode': 'w'}
    else:
        dic_ = {'stream': _sys.stdout}

    level = _log.DEBUG if debug else _log.INFO
    fmt = ('%(levelname)7s | %(asctime)s | ' +
           '%(module)15s.%(funcName)-20s[%(lineno)4d] ::: %(message)s')
    _log.basicConfig(format=fmt, datefmt='%F %T', level=level, **dic_)


def beam_rigidity(energy):
    """Return beam rigidity, beta amd game, given its energy [GeV]."""
    brho, _, beta, gamma, _ = _beam.beam_rigidity(energy=energy)
    return brho, beta, gamma


def check_pv_online(pvname, timeout=1.0, use_prefix=True):
    """Return whether a PV is online."""
    if use_prefix:
        pref = _envars.VACA_PREFIX
        pvname = pref + ('-' if pref else '') + pvname
    pvobj = _epics.PV(pvname=pvname, connection_timeout=timeout)
    status = pvobj.wait_for_connection(timeout=timeout)
    # invoke pv disconnect and explicitly signal to GC that PV object
    # may be collected.
    del pvobj
    return status


def get_strength_label(magfunc):
    """Return magnetic function strength label."""
    if magfunc == 'dipole':
        return 'Energy'
    elif magfunc in ('quadrupole', 'quadrupole-skew'):
        return 'KL'
    elif magfunc in ('sextupole',):
        return 'SL'
    elif magfunc in ('corrector-horizontal', 'corrector-vertical'):
        return 'Kick'
    else:
        raise ValueError('magfunc "{}" not defined!'.format(magfunc))


def get_strength_units(magfunc, pstype=None, pulsed=False):
    """Return strength units."""
    if magfunc == 'dipole':
        return 'GeV'
    elif magfunc in ('quadrupole', 'quadrupole-skew'):
        return '1/m'
    elif magfunc in ('sextupole',):
        return '1/m^2'
    elif magfunc in ('corrector-horizontal', 'corrector-vertical'):
        if pulsed:
            return 'mrad'
        if pstype and pstype.startswith('li-spect'):
            return 'deg'
        return 'urad'
    else:
        raise ValueError('magfunc "{}" not defined!'.format(magfunc))


def update_bit(v, bit_pos, bit_val):
    """Update a specific integer bit.

    Parameters
    ----------
    v: non-negative int
        Integer whose bit will be updated.
    bit_pos: non-negative int
        The position of the bit to update.
    bit_val: Any python object that evaluates to True or False
        Value to put on the bit.

    """
    if not isinstance(v, int) or not isinstance(bit_pos, int):
        raise TypeError("v and bit_pos must be integers.")

    if v < 0 or bit_pos < 0:
        raise ValueError("v and bit_pos must be non-negative.")

    return v | (1 << bit_pos) if bit_val else v & ~(1 << bit_pos)


def get_bit(v, bit_pos):
    """Update a specific integer bit.

    Parameters
    ----------
    v: non-negative int
        Integer whose bit will be get.
    bit_pos: int
        The position of the bit to get.

    """
    if not isinstance(v, int) or not isinstance(bit_pos, int):
        raise TypeError("v and bit_pos must be integers")

    if v < 0 or bit_pos < 0:
        raise ValueError("v and bit_pos must be non-negative.")

    return (v >> bit_pos) & 1


def mode(lst):
    """Find statistical mode of iterable of hashable objects."""
    counter = _Counter(lst)
    return counter.most_common(1)[0]


def check_public_interface_namespace(namespace, valid_interface,
                                     checkdoc_flag=True,
                                     print_flag=True):
    """Check function used in unittests to test module's public interface.

    This function checks only static public interface symbols. It does not
    check those symbols that are created within class methods.
    """
    for name in namespace.__dict__:
        if checkdoc_flag:
            doc = getattr(name, '__doc__')
            if doc is None or len(doc) < 5:
                if print_flag:
                    print('"' + name + '" has an invalid docstring!')
                return False
        if not name.startswith('_') and name not in valid_interface:
            if print_flag:
                print('Invalid symbol: ', name)
            return False
    for name in valid_interface:
        if name not in namespace.__dict__:
            if print_flag:
                print('Missing symbol: ', name)
            return False
    return True


# This solution was copied from:
# https://stackoverflow.com/questions/5189699/how-to-make-a-class-property
class ClassProperty:
    """This is a way of defining a readable class property.

    It is useful to delay the process of reading static tables during
    modules initialization (import time)
    """

    def __init__(self, func):
        """."""
        self.func = func

    def __get__(self, obj, owner):
        """."""
        return self.func(owner)



"""IOC Search module."""

import re as _re
from threading import Lock as _Lock
import yaml as _yaml

from .. import clientweb as _web


class IOCSearch:
    """IOC Search Class."""

    _service_list = list()
    _ioc_list = list()
    _service_2_ioc_dict = dict()
    _service_2_prefix_dict = dict()
    _ioc_2_service_dict = dict()
    _ioc_2_prefix_dict = dict()

    _lock = _Lock()

    @staticmethod
    def get_services(pattern='.*'):
        """Return a sorted and filtered list of services."""
        IOCSearch._reload_service_2_ioc_dict()
        pattern = _re.compile(pattern)
        auxlist = list()
        for serv in IOCSearch._service_list:
            if pattern.match(serv):
                auxlist.append(serv)
        return auxlist

    @staticmethod
    def get_iocs(pattern='.*'):
        """Return a sorted and filtered list of IOCs."""
        IOCSearch._reload_service_2_ioc_dict()
        pattern = _re.compile(pattern)
        auxlist = list()
        for ioc in IOCSearch._ioc_list:
            if pattern.match(ioc):
                auxlist.append(ioc)
        return auxlist

    @staticmethod
    def conv_service_2_ioc(service):
        """Return IOCs associated with a service."""
        IOCSearch._reload_service_2_ioc_dict()
        return IOCSearch._service_2_ioc_dict[service]

    @staticmethod
    def conv_ioc_2_service(ioc):
        """Return service associated with an IOC."""
        IOCSearch._reload_service_2_ioc_dict()
        return IOCSearch._ioc_2_service_dict[ioc]

    @staticmethod
    def conv_ioc_2_prefix(ioc):
        """Return PV prefixes associated with a ioc."""
        IOCSearch._reload_service_2_ioc_dict()
        return IOCSearch._ioc_2_prefix_dict[ioc]

    @staticmethod
    def conv_service_2_prefix(service):
        """Return PV prefixes associated with a service."""
        IOCSearch._reload_service_2_ioc_dict()
        iocs = IOCSearch.conv_service_2_ioc(service)
        prefixes = list()
        for ioc in iocs:
            pref = IOCSearch._ioc_2_prefix_dict[ioc]
            prefixes.extend(pref)
        return prefixes

    @staticmethod
    def conv_prefix_2_ioc(prefix_pattern):
        """Return IOCs that match a prefix pattern."""
        IOCSearch._reload_service_2_ioc_dict()
        pattern = _re.compile(prefix_pattern)
        iocs = set()
        for ioc in IOCSearch._ioc_list:
            prefixes = IOCSearch._ioc_2_prefix_dict[ioc]
            for pref in prefixes:
                if pattern.match(pref):
                    iocs.add(ioc)
        return iocs

    @staticmethod
    def conv_prefix_2_service(prefix_pattern):
        """Return services that match a prefix pattern."""
        IOCSearch._reload_service_2_ioc_dict()
        iocs = IOCSearch.conv_prefix_2_ioc(prefix_pattern)
        services = set()
        for ioc in iocs:
            serv = IOCSearch.conv_ioc_2_service(ioc)
            services.add(serv)
        return services

    @staticmethod
    def _reload_service_2_ioc_dict():
        """Load service to IOC data to a dict."""
        with IOCSearch._lock:
            if IOCSearch._service_2_ioc_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read service to IOC table from web server')
            text = _web.doc_services_read()
            data = _yaml.load(text)
            service_2_ioc_dict = dict()
            ioc_2_prefix_dict = dict()
            for service, ioc2pref in data.items():
                service_2_ioc_dict[service] = [ioc for ioc in ioc2pref]
                ioc_2_prefix_dict.update(ioc2pref)
            IOCSearch._service_2_ioc_dict = service_2_ioc_dict
            IOCSearch._ioc_2_prefix_dict = ioc_2_prefix_dict
            IOCSearch._service_list = sorted(list(service_2_ioc_dict.keys()))
            IOCSearch._ioc_list = sorted(list(ioc_2_prefix_dict.keys()))

            ioc_2_service_dict = dict()
            for serv, iocs in IOCSearch._service_2_ioc_dict.items():
                ioc_2_service_dict.update({i: serv for i in iocs})
            IOCSearch._ioc_2_service_dict = ioc_2_service_dict




import time as _time

from ..namesys import SiriusPVName
from .device import ProptyDevice as _ProptyDevice


class AFCPhysicalTrigger(_ProptyDevice):
    """AFC Physical Trigger device."""

    _properties = (
        'Dir-Sel', 'Dir-Sts',
        'DirPol-Sel', 'DirPol-Sts',
        'RcvCnt-Mon',
        'RcvLen-SP', 'RcvLen-RB',
        'TrnCnt-Mon',
        'TrnLen-SP', 'TrnLen-RB',
    )

    def __init__(self, devname, index):
        """Init."""
        if not 0 <= int(index) <= 7:
            raise NotImplementedError(index)

        propties = AFCPhysicalTrigger._properties
        # handle FOFB and BPM IOC differences
        # TODO: remove when new BPM IOC is updated.
        if SiriusPVName(devname).dev == 'BPM':
            propties += ('RcvCntRst-SP', 'TrnCntRst-SP')
        else:
            propties += ('RcvCntRst-Cmd', 'TrnCntRst-Cmd')

        super().__init__(
            devname, 'TRIGGER'+str(index), properties=propties)

    @property
    def direction(self):
        """Direction."""
        return self['Dir-Sts']

    @direction.setter
    def direction(self, value):
        self['Dir-Sel'] = value

    @property
    def polarity(self):
        """Polarity."""
        return self['DirPol-Sts']

    @polarity.setter
    def polarity(self, value):
        self['DirPol-Sel'] = value

    @property
    def receiver_counter(self):
        """Receiver counter."""
        return self['RcvCnt-Mon']

    def cmd_reset_receiver_counter(self):
        """Reset receiver counter."""
        # handle FOFB and BPM IOC differences
        # TODO: remove when new BPM IOC is updated.
        suf = 'SP' if 'BPM' in self.devname else 'Cmd'
        self[f'RcvCntRst-{suf}'] = 1
        _time.sleep(1)
        self[f'RcvCntRst-{suf}'] = 0

    @property
    def receiver_length(self):
        """Receiver length."""
        return self['RcvLen-RB']

    @receiver_length.setter
    def receiver_length(self, value):
        self['RcvLen-SP'] = value

    @property
    def transmitter_counter(self):
        """Transmitter counter."""
        return self['TrnCnt-Mon']

    def cmd_reset_transmitter_counter(self):
        """Reset transmitter counter."""
        # handle FOFB and BPM IOC differences
        # TODO: remove when new BPM IOC is updated.
        suf = 'SP' if 'BPM' in self.devname else 'Cmd'
        self[f'TrnCntRst-{suf}'] = 1
        _time.sleep(1)
        self[f'TrnCntRst-{suf}'] = 0

    @property
    def transmitter_length(self):
        """Transmitter length."""
        return self['TrnLen-RB']

    @transmitter_length.setter
    def transmitter_length(self, value):
        self['TrnLen-SP'] = value


class AFCACQLogicalTrigger(_ProptyDevice):
    """AFC ACQ Logical Trigger device."""

    _properties = (
        'RcvSrc-Sel', 'RcvSrc-Sts',
        'RcvInSel-SP', 'RcvInSel-RB',
        'TrnSrc-Sel', 'TrnSrc-Sts',
        'TrnOutSel-SP', 'TrnOutSel-RB',
    )

    def __init__(self, devname, index, acqcore=''):
        """Init."""
        if not 0 <= int(index) <= 23:
            raise NotImplementedError(index)
        propty_prefix = 'TRIGGER'+('_'+acqcore if acqcore else '')+str(index)
        super().__init__(
            devname, propty_prefix,
            properties=AFCACQLogicalTrigger._properties)

    @property
    def receiver_source(self):
        """Receiver source."""
        return self['RcvSrc-Sts']

    @receiver_source.setter
    def receiver_source(self, value):
        self['RcvSrc-Sel'] = value

    @property
    def receiver_in_sel(self):
        """Receiver in selection."""
        return self['RcvInSel-RB']

    @receiver_in_sel.setter
    def receiver_in_sel(self, value):
        self['RcvInSel-SP'] = value

    @property
    def transmitter_source(self):
        """Transmitter source."""
        return self['TrnSrc-Sts']

    @transmitter_source.setter
    def transmitter_source(self, value):
        self['TrnSrc-Sel'] = value

    @property
    def transmitter_out_sel(self):
        """Transmitter out selection."""
        return self['TrnOutSel-RB']

    @transmitter_out_sel.setter
    def transmitter_out_sel(self, value):
        self['TrnOutSel-SP'] = value



#!/usr/bin/env python-sirius

"""Unittest module for factory.py."""

from unittest import mock, TestCase
import siriuspy.util as util
import siriuspy.magnet.factory as factory
from siriuspy.magnet.factory import NormalizerFactory
from siriuspy.magnet.normalizer \
    import DipoleNormalizer, MagnetNormalizer, TrimNormalizer


PUB_INTERFACE = (
    'NormalizerFactory',
)


# @mock.patch('siriuspy.factory._MAData', autospec=True)
class TestMagnetFactory(TestCase):
    """Test MagnetFactory."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(factory, PUB_INTERFACE)
        self.assertTrue(valid)

    @mock.patch('siriuspy.magnet.factory._norm.DipoleNormalizer', autospec=True)
    def test_dipole_creation(self, mock_data):
        """Test Factory.create. dipole creation."""
        maname = 'SI-Fam:MA-B1B2'
        mock_data.return_value.magfunc.return_value = 'dipole'
        magnet = NormalizerFactory.create(maname=maname)
        self.assertIsInstance(magnet, DipoleNormalizer)

    @mock.patch('siriuspy.magnet.factory._norm.MagnetNormalizer', autospec=True)
    def test_magnet_creation(self, mock_data):
        """Test Factory.create magnet creation."""
        maname = 'SI-Fam:MA-QDA'
        mock_data.return_value.magfunc.return_value = 'quadrupole'
        magnet = NormalizerFactory.create(maname=maname)
        self.assertIsInstance(magnet, MagnetNormalizer)

    @mock.patch('siriuspy.magnet.factory._norm.MagnetNormalizer', autospec=True)
    def test_pulsed_magnet_creation(self, mock_data):
        """Test Factory.create magnet creation."""
        maname = 'SI-01SA:PM-InjDpKckr'
        mock_data.return_value.magfunc.return_value = 'corrector-vertical'
        magnet = NormalizerFactory.create(maname=maname)
        self.assertIsInstance(magnet, MagnetNormalizer)

    @mock.patch('siriuspy.magnet.factory._norm.TrimNormalizer', autospec=True)
    def test_trim_creation(self, mock_data):
        """Test Factory.create trim creation."""
        maname = 'SI-01M1:MA-QDA'
        mock_data.return_value.magfunc.return_value = 'quadrupole'
        magnet = NormalizerFactory.create(maname=maname)
        self.assertIsInstance(magnet, TrimNormalizer)



#!/usr/bin/env python-sirius

"""Unittest module for enumtypes.py."""

from unittest import mock, TestCase
import siriuspy.pwrsupply.csdev as csdev
import siriuspy.util as util


_MOCK_FLAG = True


PUB_INTERFACE = (
    'MAX_WFMSIZE_FBP',
    'DEF_WFMSIZE_FBP',
    'DEFAULT_WFM_FBP',
    'MAX_WFMSIZE_OTHERS',
    'DEF_WFMSIZE_OTHERS',
    'DEFAULT_WFM_OTHERS',
    'DEFAULT_WFM',
    'PSSOFB_MAX_NR_UDC',
    'DEFAULT_SIGGEN_CONFIG',
    'PS_CURRENT_PRECISION',
    'PU_VOLTAGE_PRECISION',
    'PS_CURRENT_UNIT',
    'PU_VOLTAGE_UNIT',
    'PS_LI_INTLK_THRS',
    'ETypes',
    'Const',
    'get_ps_propty_database',
    'get_conv_propty_database',
    'get_ps_interlocks',
    'get_ps_modules',
    'get_ps_scopesourcemap',
)


class TestPwrSupplyCSDev(TestCase):
    """Test pwrsupply csdev module."""

    ps_alarm = ('Current-SP', 'Current-RB',
                'CurrentRef-Mon', 'Current-Mon', )
    pu_alarm = ('Voltage-SP', 'Voltage-RB',
                'Voltage-Mon', )
    pstypes = [
        'si-dipole-b1b2-fam',
        'si-quadrupole-q14-fam',
        'si-sextupole-s15-sd-fam',
        'si-sextupole-s15-ch',
        'si-sextupole-s15-cv',
        'si-sextupole-s15-qs',
        'si-corrector-fcv',
        'bo-ejekicker',
        'si-hping',
        'si-injdpk',
        'si-injnlk',
        'si-vping',
        'tb-injseptum',
        'ts-ejeseptum-thin',
        'ts-ejeseptum-thick',
        'ts-injseptum-thin',
        'ts-injseptum-thick',
    ]
    sample = {
        'SI-Fam:MA-B1B2': 'Energy-SP',
        'SI-Fam:MA-QDA': 'KL-SP',
        'SI-Fam:MA-SDB2': 'SL-SP',
        'SI-01C1:MA-CH': 'Kick-SP',
        'SI-02C3:MA-CV-2': 'Kick-SP',
        'SI-03M1:MA-QS': 'KL-SP',
        'SI-02M1:MA-FCV': 'Kick-SP',
        'BO-48D:PM-EjeKckr': 'Kick-SP',
        'SI-01SA:PM-PingH': 'Kick-SP',
        'SI-01SA:PM-InjDpKckr': 'Kick-SP',
        'SI-01SA:PM-InjNLKckr': 'Kick-SP',
        'SI-19C4:PM-PingV': 'Kick-SP',
        'TB-04:PM-InjSept': 'Kick-SP',
        'TS-01:PM-EjeSeptF': 'Kick-SP',
        'TS-01:PM-EjeSeptG': 'Kick-SP',
        'TS-04:PM-InjSeptF': 'Kick-SP',
        'TS-04:PM-InjSeptG-1': 'Kick-SP',
        'TS-04:PM-InjSeptG-2': 'Kick-SP',
    }

    def setUp(self):
        """Define setup method."""
        def get_splims(pstype, alarm):
            dbase = {'lolo': 0.0, 'low': 1.0, 'lolim': 2.0, 'hilim': 3.0,
                     'high': 4.0, 'hihi': 5.0}
            return dbase[alarm]

        def get_splims_unit(psmodel):
            if psmodel in ('FBP', 'FAC', 'FAP', 'FAC_2S', 'FAC_2P4S'):
                return ['A', 'Ampere']
            return ['V', 'Voltage']

        if _MOCK_FLAG:
            _pssearch_patcher = mock.patch(
                'siriuspy.pwrsupply.csdev._PSSearch', autospec=True)
            self.addCleanup(_pssearch_patcher.stop)
            self.m_pssearch = _pssearch_patcher.start()
            self.m_pssearch.get_splims_unit.side_effect = get_splims_unit
            self.m_pssearch.get_pstype_names.return_value = \
                TestPwrSupplyCSDev.pstypes
            self.m_pssearch.get_splims.side_effect = get_splims
            self.m_pssearch.conv_psname_2_psmodel.return_value = 'FBP'
            self.m_pssearch.conv_pstype_2_magfunc.return_value = 'quadrupole'

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(csdev, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_MAX_WFMSIZE_FBP(self):
        """Test MAX_WFMSIZE_FBP."""
        self.assertIsInstance(csdev.MAX_WFMSIZE_FBP, int)
        self.assertTrue(csdev.MAX_WFMSIZE_FBP > 0)

    def test_DEF_WFMSIZE_FBP(self):
        """Test DEF_WFMSIZE_FBP."""
        self.assertIsInstance(csdev.DEF_WFMSIZE_FBP, int)
        self.assertTrue(csdev.DEF_WFMSIZE_FBP > 0)
        self.assertTrue(csdev.DEF_WFMSIZE_FBP <= csdev.MAX_WFMSIZE_FBP)

    def test_MAX_WFMSIZE_OTHERS(self):
        """Test MAX_WFMSIZE_OTHERS."""
        self.assertIsInstance(csdev.MAX_WFMSIZE_OTHERS, int)
        self.assertTrue(csdev.MAX_WFMSIZE_OTHERS > 0)

    def test_DEF_WFMSIZE_OTHERS(self):
        """Test DEF_WFMSIZE_OTHERS."""
        self.assertIsInstance(csdev.DEF_WFMSIZE_OTHERS, int)
        self.assertTrue(csdev.DEF_WFMSIZE_OTHERS > 0)
        self.assertTrue(
            csdev.DEF_WFMSIZE_OTHERS <= csdev.MAX_WFMSIZE_OTHERS)

    def test_DEFAULT_SIGGEN_CONFIG(self):
        """Test DEFAULT_SIGGEN_CONFIG."""
        self.assertIsInstance(csdev.DEFAULT_SIGGEN_CONFIG, tuple)
        self.assertTrue(len(csdev.DEFAULT_SIGGEN_CONFIG), 9)

    def test_ps_propty_database(self):
        """Test ps_propty_database."""
        current_pvs = TestPwrSupplyCSDev.ps_alarm
        for pstype in TestPwrSupplyCSDev.pstypes:
            dbase = csdev.get_ps_propty_database('FBP', pstype)
            unit = dbase['Current-SP']['unit']
            for propty, dbi in dbase.items():
                # set setpoint limits in database
                if propty in TestPwrSupplyCSDev.ps_alarm:
                    self.assertLessEqual(dbi['lolo'], dbi['low'])
                    self.assertLessEqual(dbi['low'], dbi['lolim'])
                    self.assertLessEqual(dbi['lolim'], dbi['hilim'])
                    self.assertLessEqual(dbi['hilim'], dbi['high'])
                    self.assertLessEqual(dbi['high'], dbi['hihi'])
                if propty in current_pvs:
                    self.assertEqual(dbi['unit'], unit)

    def test_ps_FBP_propty_database(self):
        """Test ps_FBP_propty_database."""
        dbase = csdev.get_ps_propty_database('FBP', 'si-quadrupole-q14-fam')
        self.assertIsInstance(dbase, dict)
        for prop in dbase:
            self.assertIsInstance(dbase[prop], dict)
        # test precision consistency
        proptys = TestPwrSupplyCSDev.ps_alarm
        for propty in proptys:
            self.assertIn(propty, dbase)

    def test_ps_basic_propty_database(self):
        """Test ps_basic_propty_database."""
        dbase = csdev._get_ps_basic_propty_database()
        self.assertIsInstance(dbase, dict)
        for prop in dbase:
            self.assertIsInstance(dbase[prop], dict)

    def test_ps_common_propty_database(self):
        """Test ps_common_propty_database."""
        dbase = csdev._get_ps_common_propty_database()
        self.assertIsInstance(dbase, dict)
        for prop in dbase:
            self.assertIsInstance(dbase[prop], dict)



"""E-Gun devices."""

import time as _time
from threading import Event as _Flag
from functools import partial as _part

import numpy as _np

from ..pwrsupply.psctrl.pscstatus import PSCStatus as _PSCStatus
from ..injctrl.csdev import Const as _InjConst
from ..callbacks import Callback as _Callback

from .device import Device as _Device, Devices as _Devices, \
    DeviceNC as _DeviceNC
from .timing import Trigger


class EGBias(_Device):
    """EGun Bias Device."""

    DEF_TIMEOUT = 10  # [s]
    PWRSTATE = _PSCStatus.PWRSTATE

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-BiasPS'
        ALL = (LI, )

    _properties = (
        'voltoutsoft', 'voltinsoft', 'currentinsoft', 'switch', 'swstatus')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = EGBias.DEVICES.LI

        # check if device exists
        if devname not in EGBias.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGBias._properties)

    @property
    def voltage_mon(self):
        """."""
        return self['voltinsoft']

    @property
    def voltage(self):
        """."""
        return self['voltoutsoft']

    @voltage.setter
    def voltage(self, value):
        self['voltoutsoft'] = value

    @property
    def current_mon(self):
        """."""
        return self['currentinsoft']

    def cmd_turn_on(self, timeout=DEF_TIMEOUT):
        """."""
        self['switch'] = self.PWRSTATE.On
        return self._wait('swstatus', self.PWRSTATE.On, timeout)

    def cmd_turn_off(self, timeout=DEF_TIMEOUT):
        """."""
        self['switch'] = self.PWRSTATE.Off
        return self._wait('swstatus', self.PWRSTATE.Off, timeout)

    def is_on(self):
        """."""
        return self['swstatus'] == self.PWRSTATE.On

    def set_voltage(self, value, tol=0.2, timeout=DEF_TIMEOUT):
        """Set voltage and wait readback reach value with a tolerance."""
        self.voltage = value
        nrp = int(timeout / 0.1)
        for _ in range(nrp):
            if abs(self.voltage_mon - value) < tol:
                return True
            _time.sleep(0.1)
        print('timed out waiting for EGBias voltage to reach ',
              value, 'with tolerance ', tol)
        return False


class EGFilament(_Device):
    """EGun Filament Device."""

    DEF_TIMEOUT = 10  # [s]
    PWRSTATE = _PSCStatus.PWRSTATE

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-FilaPS'
        ALL = (LI, )

    _properties = (
        'voltinsoft', 'currentinsoft', 'currentoutsoft', 'switch', 'swstatus')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = EGFilament.DEVICES.LI

        # check if device exists
        if devname not in EGFilament.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGFilament._properties)

    @property
    def voltage_mon(self):
        """."""
        return self['voltinsoft']

    @property
    def current_mon(self):
        """."""
        return self['currentinsoft']

    @property
    def current(self):
        """."""
        return self['currentoutsoft']

    @current.setter
    def current(self, value):
        """."""
        self['currentoutsoft'] = value

    def cmd_turn_on(self, timeout=DEF_TIMEOUT):
        """."""
        self['switch'] = self.PWRSTATE.On
        return self._wait('swstatus', self.PWRSTATE.On, timeout)

    def cmd_turn_off(self, timeout=DEF_TIMEOUT):
        """."""
        self['switch'] = self.PWRSTATE.Off
        return self._wait('swstatus', self.PWRSTATE.Off, timeout)

    def is_on(self):
        """."""
        return self['swstatus'] == self.PWRSTATE.On

    def wait_current(self, value, tol=0.20, timeout=DEF_TIMEOUT):
        """Wait current to reach value with tolerance 'tol'."""
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if abs(self.current_mon - value) < tol:
                return True
            _time.sleep(0.1)
        return False


class EGHVPS(_Device):
    """Egun High-Voltage Power Supply Device."""

    DEF_TIMEOUT = 10  # [s]
    PWRSTATE = _PSCStatus.PWRSTATE

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-HVPS'
        ALL = (LI, )

    _properties = (
        'currentinsoft', 'currentoutsoft',
        'voltinsoft', 'voltoutsoft',
        'enable', 'enstatus',
        'switch', 'swstatus')

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = EGHVPS.DEVICES.LI

        # check if device exists
        if devname not in EGHVPS.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGHVPS._properties)

    @property
    def current_mon(self):
        """."""
        return self['currentinsoft']

    @property
    def current(self):
        """."""
        return self['currentoutsoft']

    @current.setter
    def current(self, value):
        """."""
        self['currentoutsoft'] = value

    @property
    def voltage_mon(self):
        """."""
        return self['voltinsoft']

    @property
    def voltage(self):
        """."""
        return self['voltoutsoft']

    @voltage.setter
    def voltage(self, value):
        self['voltoutsoft'] = value

    def cmd_enable(self, timeout=DEF_TIMEOUT):
        """Enable."""
        if self['enstatus'] == self.PWRSTATE.Off:
            self['enable'] = self.PWRSTATE.Off
            _time.sleep(1)

        self['enable'] = self.PWRSTATE.On
        if not self._wait('enstatus', self.PWRSTATE.On, timeout=timeout):
            return False
        return True

    def cmd_turn_on(self, timeout=DEF_TIMEOUT):
        """."""
        self['enable'] = self.PWRSTATE.On
        if not self._wait('enstatus', self.PWRSTATE.On, timeout=timeout/2):
            return False
        self['switch'] = self.PWRSTATE.On
        return self._wait('swstatus', self.PWRSTATE.On, timeout=timeout/2)

    def cmd_turn_off(self, timeout=DEF_TIMEOUT):
        """."""
        self['enable'] = self.PWRSTATE.Off
        if not self._wait('enstatus', self.PWRSTATE.Off, timeout=timeout/2):
            return False
        self['switch'] = self.PWRSTATE.Off
        return self._wait('swstatus', self.PWRSTATE.Off, timeout=timeout/2)

    def is_on(self):
        """."""
        ison = self['enstatus'] == self.PWRSTATE.On
        ison &= self['swstatus'] == self.PWRSTATE.On
        return ison

    def wait_voltage(self, value, tol=1, timeout=DEF_TIMEOUT):
        """Wait voltage to reach value with tolerance 'tol'."""
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if abs(self.voltage_mon - value) < tol:
                return True
            _time.sleep(0.1)
        return False

    def wait_current(self, value, tol=1e-3, timeout=DEF_TIMEOUT):
        """Wait current setpoint to reach value with tolerance 'tol'."""
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if abs(self.current - value) < tol:
                return True
            _time.sleep(0.1)
        return False


class EGTriggerPS(_Device):
    """Egun Trigger Power Supply Device."""

    DEF_TIMEOUT = 10  # [s]

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-TriggerPS'
        ALL = (LI, )

    _properties = (
        'status', 'allow', 'enable', 'enablereal')

    def __init__(self, devname=DEVICES.LI):
        """."""
        # check if device exists
        if devname not in EGTriggerPS.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGTriggerPS._properties)

    @property
    def status(self):
        """."""
        return self['status']

    @property
    def allow(self):
        """."""
        return self['allow']

    @property
    def enable(self):
        """."""
        return self['enablereal']

    @enable.setter
    def enable(self, value):
        self['enable'] = bool(value)

    def cmd_enable_trigger(self, timeout=DEF_TIMEOUT):
        """."""
        self['enable'] = 1
        return self._wait('enablereal', value=1, timeout=timeout)

    def cmd_disable_trigger(self, timeout=DEF_TIMEOUT):
        """."""
        self['enable'] = 0
        return self._wait('enablereal', value=0, timeout=timeout)

    def is_on(self):
        """."""
        return self['enablereal'] == 1


class EGPulsePS(_Device):
    """Egun Pulse Power Supply Device."""

    DEF_TIMEOUT = 10  # [s]

    class DEVICES:
        """Devices names."""

        LI = 'LI-01:EG-PulsePS'
        ALL = (LI, )

    _properties = (
        'singleselect', 'singleselstatus', 'singleswitch', 'singleswstatus',
        'multiselect', 'multiselstatus', 'multiswitch', 'multiswstatus',
        'poweroutsoft', 'powerinsoft')

    def __init__(self, devname=DEVICES.LI):
        """Init."""
        # check if device exists
        if devname not in EGPulsePS.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EGPulsePS._properties)

    @property
    def single_bunch_mode(self):
        """Single bunch mode."""
        return self['singleselstatus']

    @single_bunch_mode.setter
    def single_bunch_mode(self, value):
        self['singleselect'] = value

    @property
    def single_bunch_switch(self):
        """Single bunch switch."""
        return self['singleswstatus']

    @single_bunch_switch.setter
    def single_bunch_switch(self, value):
        self['singleswitch'] = value

    @property
    def multi_bunch_mode(self):
        """Multi bunch mode."""
        return self['multiselstatus']

    @multi_bunch_mode.setter
    def multi_bunch_mode(self, value):
        self['multiselect'] = value

    @property
    def multi_bunch_switch(self):
        """Multi bunch switch."""
        return self['multiswstatus']

    @multi_bunch_switch.setter
    def multi_bunch_switch(self, value):
        self['multiswitch'] = value

    @property
    def power_mon(self):
        """Power monitor."""
        return self['powerinsoft']

    @property
    def power(self):
        """Power readback."""
        return self['poweroutsoft']

    @power.setter
    def power(self, value):
        self['poweroutsoft'] = value

    def cmd_turn_off_single_bunch_switch(self, timeout=DEF_TIMEOUT):
        """Turn off single bunch switch."""
        self.single_bunch_switch = 0
        return self._wait('singleswstatus', 0, timeout=timeout)

    def cmd_turn_off_single_bunch_mode(self, timeout=DEF_TIMEOUT):
        """Turn off single bunch mode."""
        self.single_bunch_mode = 0
        return self._wait('singleselstatus', 0, timeout=timeout)

    def cmd_turn_off_single_bunch(self, timeout=DEF_TIMEOUT):
        """Turn off single bunch."""
        if not self.cmd_turn_off_single_bunch_switch():
            return False
        return self.cmd_turn_off_single_bunch_mode()

    def cmd_turn_on_single_bunch_switch(self, timeout=DEF_TIMEOUT):
        """Turn on single bunch switch."""
        self.single_bunch_switch = 1
        return self._wait('singleswstatus', 1, timeout=timeout)

    def cmd_turn_on_single_bunch_mode(self, timeout=DEF_TIMEOUT):
        """Turn on single bunch mode."""
        self.single_bunch_mode = 1
        return self._wait('singleselstatus', 1, timeout=timeout)

    def cmd_turn_on_single_bunch(self, timeout=DEF_TIMEOUT):
        """Turn on single bunch."""
        if not self.cmd_turn_on_single_bunch_mode():
            return False
        return self.cmd_turn_on_single_bunch_switch()

    def cmd_turn_off_multi_bunch_switch(self, timeout=DEF_TIMEOUT):
        """Turn off multi bunch switch."""
        self.multi_bunch_switch = 0
        return self._wait('multiswstatus', 0, timeout=timeout)

    def cmd_turn_off_multi_bunch_mode(self, timeout=DEF_TIMEOUT):
        """Turn off multi bunch mode."""
        self.multi_bunch_mode = 0
        return self._wait('multiselstatus', 0, timeout=timeout)

    def cmd_turn_off_multi_bunch(self, timeout=DEF_TIMEOUT):
        """Turn off multi bunch."""
        if not self.cmd_turn_off_multi_bunch_switch():
            return False
        return self.cmd_turn_off_multi_bunch_mode()

    def cmd_turn_on_multi_bunch_switch(self, timeout=DEF_TIMEOUT):
        """Turn on multi bunch switch."""
        self.multi_bunch_switch = 1
        return self._wait('multiswstatus', 1, timeout=timeout)

    def cmd_turn_on_multi_bunch_mode(self, timeout=DEF_TIMEOUT):
        """Turn on multi bunch mode."""
        self.multi_bunch_mode = 1
        return self._wait('multiselstatus', 1, timeout=timeout)

    def cmd_turn_on_multi_bunch(self, timeout=DEF_TIMEOUT):
        """Turn on multi bunch."""
        if not self.cmd_turn_on_multi_bunch_mode():
            return False
        return self.cmd_turn_on_multi_bunch_switch()


class EGun(_Devices, _Callback):
    """EGun device."""

    DEF_TIMEOUT = 10  # [s]
    BIAS_MULTI_BUNCH = _InjConst.BIAS_MULTI_BUNCH
    BIAS_SINGLE_BUNCH = _InjConst.BIAS_SINGLE_BUNCH
    BIAS_TOLERANCE = 1.0  # [V]
    HV_OPVALUE = _InjConst.HV_OPVALUE
    HV_TOLERANCE = 1.0  # [kV]
    HV_LEAKCURR_OPVALUE = 0.015  # [mA]
    HV_MAXVALUE = 90.0  # [kV]
    HV_RAMPUP_NRPTS = 15
    HV_RAMPDN_NRPTS = 6
    HV_RAMP_DURATION = 70  # [s]
    FILACURR_OPVALUE = _InjConst.FILACURR_OPVALUE
    FILACURR_TOLERANCE = 0.20  # [A]
    FILACURR_MAXVALUE = 1.42  # [A]
    FILACURR_RAMP_NRPTS = 10
    FILACURR_RAMP_DURATION = 7*60  # [s]

    def __init__(self, print_log=True, callback=None):
        """Init."""
        self.bias = EGBias()
        self.fila = EGFilament()
        self.hvps = EGHVPS()
        self.trigps = EGTriggerPS()
        self.pulse = EGPulsePS()
        self.trigsingle = Trigger('LI-01:TI-EGun-SglBun')
        self.trigmulti = Trigger('LI-01:TI-EGun-MultBun')
        self.trigmultipre = Trigger('LI-01:TI-EGun-MultBunPre')

        self.sys_ext = _DeviceNC('LI-01:EG-External', ('status', ))
        self.sys_vlv = _DeviceNC('LI-01:EG-Valve', ('status', ))
        self.sys_gat = _DeviceNC('LI-01:EG-Gate', ('status', ))
        self.sys_vac = _DeviceNC('LI-01:EG-Vacuum', ('status', ))
        self.mps_ccg = _DeviceNC(
            'LA-CN:H1MPS-1', ('CCG1Warn_L', 'CCG2Warn_L'))
        self.mps_gun = _DeviceNC('LA-CN:H1MPS-1', ('GunPermit', ))

        devices = (
            self.bias, self.fila, self.hvps, self.trigps, self.pulse,
            self.trigsingle, self.trigmulti, self.trigmultipre,
            self.sys_ext, self.sys_vlv, self.sys_gat, self.sys_vac,
            self.mps_ccg, self.mps_gun)

        self._bias_mb = EGun.BIAS_MULTI_BUNCH
        self._bias_sb = EGun.BIAS_SINGLE_BUNCH
        self._bias_tol = EGun.BIAS_TOLERANCE
        self._hv_opval = EGun.HV_OPVALUE
        self._hv_tol = EGun.HV_TOLERANCE
        self._hv_leakcurr = EGun.HV_LEAKCURR_OPVALUE
        self._filacurr_opval = EGun.FILACURR_OPVALUE
        self._filacurr_tol = EGun.FILACURR_TOLERANCE
        self._print_log = print_log
        self._abort_chg_type = _Flag()
        self._abort_rmp_hvps = _Flag()
        self._abort_rmp_fila = _Flag()

        _Devices.__init__(self, '', devices)
        _Callback.__init__(self, callback=callback)

    @property
    def multi_bunch_bias_voltage(self):
        """Multi bunch bias voltage to be used in mode setup."""
        return self._bias_mb

    @multi_bunch_bias_voltage.setter
    def multi_bunch_bias_voltage(self, value):
        self._bias_mb = value

    @property
    def single_bunch_bias_voltage(self):
        """Single bunch bias voltage to be used in mode setup."""
        return self._bias_sb

    @single_bunch_bias_voltage.setter
    def single_bunch_bias_voltage(self, value):
        self._bias_sb = value

    @property
    def bias_voltage_tol(self):
        """Bias voltage tolerance."""
        return self._bias_tol

    @bias_voltage_tol.setter
    def bias_voltage_tol(self, value):
        self._bias_tol = value

    def cmd_switch_to_single_bunch(self):
        """Switch to single bunch mode."""
        self._abort_chg_type.clear()
        self._update_status('Switching EGun mode to SingleBunch...')

        if not self.wait_for_connection(1):
            self._update_status('ERR:EGun device not connected. Aborted.')
            return False

        # fun, msg
        proced = (
            (_part(self.trigps.cmd_disable_trigger), 'disable EGun Trigger.'),
            (_part(self.pulse.cmd_turn_off_multi_bunch),
             'turn off MultiBunch.'),
            (_part(
                self.bias.set_voltage, self._bias_sb, tol=self._bias_tol),
             'set EGun Bias voltage to SB operation value.'),
            (_part(self.trigmultipre.cmd_disable), 'disable MB pre trigger.'),
            (_part(self.trigmulti.cmd_disable), 'disable MB trigger.'),
            (_part(self.trigsingle.cmd_enable), 'enable SB trigger.'),
            (_part(self.pulse.cmd_turn_on_single_bunch),
             'turn on SingleBunch.'),
            )

        for fun, msg in proced:
            if self._abort_chg_type.is_set():
                self._clr_flag_abort_chg_type()
                return False
            if not fun():
                self._update_status('ERR:Could not ' + msg)
                return False
        self._update_status('EGun switched to SingleBunch!')
        return True

    def cmd_switch_to_multi_bunch(self):
        """Switch to multi bunch mode."""
        self._abort_chg_type.clear()
        self._update_status('Switching EGun mode to MultiBunch...')

        if not self.wait_for_connection(1):
            self._update_status('ERR:EGun device not connected. Aborted.')
            return False

        # fun, msg
        proced = (
            (_part(self.trigps.cmd_disable_trigger), 'disable EGun Trigger.'),
            (_part(self.pulse.cmd_turn_off_single_bunch),
             'turn off SingleBunch.'),
            (_part(self.bias.set_voltage, self._bias_mb, tol=self._bias_tol),
             'set EGun Bias voltage to MB operation value.'),
            (_part(self.trigsingle.cmd_disable), 'disable SB trigger.'),
            (_part(self.trigmultipre.cmd_enable), 'disable SB pre trigger.'),
            (_part(self.trigmulti.cmd_enable), 'enable MB trigger.'),
            (_part(self.pulse.cmd_turn_on_multi_bunch), 'turn on MultiBunch.'),
        )

        for fun, msg in proced:
            if self._abort_chg_type.is_set():
                self._clr_flag_abort_chg_type()
                return False
            if not fun():
                self._update_status('ERR:Could not ' + msg)
                return False
        self._update_status('EGun configured to MultiBunch!')
        return True

    @property
    def is_single_bunch(self):
        """Is configured to single bunch mode."""
        if not self.wait_for_connection(1):
            return False
        sts = not self.pulse.multi_bunch_switch
        sts &= not self.pulse.multi_bunch_mode
        sts &= not self.trigmultipre.state
        sts &= not self.trigmulti.state
        sts &= self.trigsingle.state
        sts &= self.pulse.single_bunch_mode
        sts &= self.pulse.single_bunch_switch
        return sts

    @property
    def is_multi_bunch(self):
        """Is configured to multi bunch mode."""
        if not self.wait_for_connection(1):
            return False
        sts = not self.pulse.single_bunch_switch
        sts &= not self.pulse.single_bunch_mode
        sts &= not self.trigsingle.state
        sts &= self.trigmultipre.state
        sts &= self.trigmulti.state
        sts &= self.pulse.multi_bunch_mode
        sts &= self.pulse.multi_bunch_switch
        return sts

    @property
    def high_voltage_opvalue(self):
        """High voltage operation value."""
        return self._hv_opval

    @high_voltage_opvalue.setter
    def high_voltage_opvalue(self, value):
        self._hv_opval = value

    @property
    def high_voltage_leakcurr(self):
        """High voltage leakage current value."""
        return self._hv_leakcurr

    @high_voltage_leakcurr.setter
    def high_voltage_leakcurr(self, value):
        self._hv_leakcurr = value
        self.hvps.current = value

    @property
    def is_hv_on(self):
        """Indicate whether high voltage is on and in operational value."""
        if not self.hvps.wait_for_connection(1):
            return False
        is_on = self.hvps.is_on()
        is_op = abs(self.hvps.voltage_mon - self._hv_opval) < self._hv_tol
        return is_on and is_op

    def set_hv_voltage(self, value=None, duration=None, timeout=DEF_TIMEOUT):
        """Set HVPS voltage."""
        self._abort_rmp_hvps.clear()
        self._update_status('Setpoint received for HVPS voltage...')
        if not self._check_status_ok():
            self._update_status('ERR:MPS or LI Status not ok. Aborted.')
            return False
        if not self.hvps['swstatus'] == self.hvps.PWRSTATE.On:
            self._update_status('ERR:HVPS switch is not on.')
            return False
        if self._abort_rmp_hvps.is_set():
            self._clr_flag_abort_rmp_hvps()
            return False
        if value is None:
            value = self._hv_opval

        # if voltage already satisfies value
        cond = abs(self.hvps['voltinsoft'] - value) < self._hv_tol
        cond &= abs(self.hvps['voltoutsoft'] - value) < 1e-4
        if cond:
            self._update_status(
                f'HVPS Voltage is already at {value:.3f}kV.')
            self.hvps.voltage = value
            return True

        if self._abort_rmp_hvps.is_set():
            self._clr_flag_abort_rmp_hvps()
            return False

        # if needed, enable HVPS
        if not self.hvps['enstatus'] == self.hvps.PWRSTATE.On:
            self._update_status('Sending enable command to HVPS...')
            if not self.hvps.cmd_enable(5):
                self._update_status('ERR:Could not enable HVPS.')
                return False
            self._update_status('HVPS enabled.')

        if self._abort_rmp_hvps.is_set():
            self._clr_flag_abort_rmp_hvps()
            return False

        # before voltage setpoints, set leakage current to suitable value
        self._update_status(
            f'Setting max. leak current to {self._hv_leakcurr:.3f}mA.')
        self.hvps.current = self._hv_leakcurr
        if not self.hvps.wait_current(self._hv_leakcurr):
            self._update_status(
                'ERR:Timed out waiting for HVPS leak current.')
            return False

        if self._abort_rmp_hvps.is_set():
            self._clr_flag_abort_rmp_hvps()
            return False

        # if value is lower, do a ramp down
        if value < self.hvps.voltage_mon:
            nrpts = EGun.HV_RAMPDN_NRPTS
            power = 1
        else:  # else, do a ramp up
            nrpts = EGun.HV_RAMPUP_NRPTS
            power = 2
        duration = duration if duration is not None else EGun.HV_RAMP_DURATION
        max_value = EGun.HV_MAXVALUE
        ydata = self._get_ramp(
            self.hvps.voltage_mon, value, nrpts, max_value, power)
        t_inter = duration / (nrpts-1)

        self._update_status(f'Starting HVPS ramp to {value:.3f}kV.')
        self._update_status(
            f'This process will take about {ydata.size*t_inter:.1f}s.')
        for i, volt in enumerate(ydata[1:]):
            if self._abort_rmp_hvps.is_set():
                self._clr_flag_abort_rmp_hvps()
                return False
            self.hvps.voltage = volt
            self._update_status(
                f'{i+1:02d}/{len(ydata[1:]):02d} -> HV = {volt:.3f}kV')
            _time.sleep(t_inter)
            _t0 = _time.time()
            while _time.time() - _t0 < timeout:
                if self._abort_rmp_hvps.is_set():
                    self._clr_flag_abort_rmp_hvps()
                    return False
                if not self._check_status_ok():
                    self._update_status(
                        'ERR:MPS or LI Status not ok. Aborted.')
                    return False
                if abs(self.hvps.voltage_mon - volt) < self._hv_tol:
                    break
                _time.sleep(0.1)
            else:
                self._update_status(
                    'ERR:HVPS Voltage is taking too '
                    'long to reach setpoint. Aborted.')
                return False
        self._update_status('HVPS Ready!')
        return True

    @property
    def fila_current_opvalue(self):
        """Filament current operation value."""
        return self._filacurr_opval

    @fila_current_opvalue.setter
    def fila_current_opvalue(self, value):
        self._filacurr_opval = value

    @property
    def is_fila_on(self):
        """Indicate whether filament is on and in operational current."""
        if not self.fila.wait_for_connection(1):
            return False
        is_on = self.fila.is_on()
        is_op_sp = abs(self.fila['currentoutsoft']-self._filacurr_opval) < 1e-4
        is_op_rb = abs(self.fila['currentinsoft']-self._filacurr_opval) < \
            self._filacurr_tol
        return is_on and is_op_sp and is_op_rb

    def set_fila_current(self, value=None):
        """Set filament current."""
        self._abort_rmp_fila.clear()
        self._update_status('Setpoint received for FilaPS current...')
        if not self._check_status_ok():
            self._update_status('ERR:MPS or LI Status not ok. Aborted.')
            return False
        if not self.fila.is_on():
            self._update_status('ERR:FilaPS is not on.')
            return False
        if self._abort_rmp_fila.is_set():
            self._clr_flag_abort_rmp_fila()
            return False
        if value is None:
            value = self._filacurr_opval

        # if current already satisfies value
        cond = abs(self.fila['currentinsoft'] - value) < self._filacurr_tol
        cond &= abs(self.fila['currentoutsoft'] - value) < 1e-4
        if cond:
            self._update_status(
                'FilaPS current is already at {0:.3f}A.'.format(value))
            self.fila.current = value
            return True

        if self._abort_rmp_fila.is_set():
            self._clr_flag_abort_rmp_fila()
            return False

        # elif value is lower, do only one setpoint
        if value < self.fila.current_mon:
            self._update_status(f'Setting current to {value:.3f}A...')
            self.fila.current = value
            if self.fila.wait_current(value, self._filacurr_tol):
                self._update_status('FilaPS Ready!')
                return True
            self._update_status(
                'ERR:Timed out waiting for FilaPS current.')
            return False

        if self._abort_rmp_fila.is_set():
            self._clr_flag_abort_rmp_fila()
            return False

        # else, do a ramp up
        duration = EGun.FILACURR_RAMP_DURATION
        nrpts = EGun.FILACURR_RAMP_NRPTS
        max_value = EGun.FILACURR_MAXVALUE
        ydata = self._get_ramp(self.fila.current_mon, value, nrpts, max_value)
        t_inter = duration / (nrpts-1)
        total_steps_duration = (len(ydata)-1)*t_inter

        self._update_status(
            f'Starting filament ramp to {value:.3f} A.')
        for i, cur in enumerate(ydata[1:]):
            if self._abort_rmp_fila.is_set():
                self._clr_flag_abort_rmp_fila()
                return False
            self.fila.current = cur
            dur = total_steps_duration - i*t_inter
            self._update_status(
                f'{i+1:02d}/{len(ydata[1:]):02d} -> '
                f'Rem. Time: {dur:03.0f}s  Curr: {cur:.3f} A')
            _time.sleep(t_inter)

            if not self._check_status_ok():
                self._update_status(
                    'ERR:MPS or LI Status not ok. Aborted.')
                return False
        self._update_status('FilaPS Ready!')
        return True

    # -------- thread help methods --------

    def cmd_abort_chg_type(self):
        """Abort injection type change."""
        self._abort_chg_type.set()
        self._update_status('WARN:Abort received for Type change.')
        return True

    def _clr_flag_abort_chg_type(self):
        """Clear abort flag for injection type change."""
        self._abort_chg_type.clear()
        self._update_status('ERR:Aborted Type change.')
        return True

    def cmd_abort_rmp_hvps(self):
        """Abort set HVPS voltage ramp."""
        self._abort_rmp_hvps.set()
        self._update_status('WARN:Abort received for HVPS ramp.')
        return True

    def _clr_flag_abort_rmp_hvps(self):
        """Clear abort flag for HVPS voltage ramp."""
        self._abort_rmp_hvps.clear()
        self._update_status('ERR:Aborted HVPS voltage ramp.')
        return True

    def cmd_abort_rmp_fila(self):
        """Abort FilaPS current ramp."""
        self._abort_rmp_fila.set()
        self._update_status('WARN:Abort received for FilaPS ramp.')
        return True

    def _clr_flag_abort_rmp_fila(self):
        """Clear abort flag for FilaPS current ramp."""
        self._abort_rmp_fila.clear()
        self._update_status('ERR:Aborted FilaPS current ramp.')
        return True

    # ---------- private methods ----------

    def _get_ramp(self, curr_val, goal, nrpts, max_val, power=2):
        xdata = _np.linspace(0, 1, nrpts)
        if curr_val > goal:
            xdata = _np.flip(xdata)

        ydata = 1 - (1-xdata)**power
        # ydata = _np.sqrt(1 - (1-xdata)**2)
        # ydata = _np.sin(2*np.pi * xdata/4)

        ydata *= max_val
        mini = min(curr_val, goal)
        maxi = max(curr_val, goal)
        ydata = ydata[ydata > mini]
        ydata = ydata[ydata < maxi]
        ydata = _np.r_[curr_val, ydata, goal]
        return ydata

    def _check_status_ok(self):
        """Check if interlock signals are ok."""
        if not self.wait_for_connection(1):
            return False
        isok = [self.mps_ccg[ppty] == 0 for ppty in self.mps_ccg.properties]
        allok = all(isok)
        allok &= self.mps_gun['GunPermit'] == 1
        allok &= self.sys_ext['status'] == 1
        allok &= self.sys_vlv['status'] == 1
        allok &= self.sys_gat['status'] == 1
        allok &= self.sys_vac['status'] == 1
        return allok

    # ---------- logging -----------

    def _update_status(self, status):
        if self._print_log:
            print(status)
        self.run_callbacks(status)



#!/usr/bin/env python-sirius

"""Test DipoleNormalizer.

Requirements:
    convert current to strength
    convert strength to current
    all sections: TB, BO, TS, SI
"""
from unittest import TestCase

from siriuspy.magnet.normalizer import DipoleNormalizer


class DipoleNormalizerTest(TestCase):
    """Test conversion of current/energy in dipoles."""

    settings = {
        "TB": {"name": "TB-Fam:MA-B",
               "current": 476.0,
               "strength": 0.28171517156419867},
        "BO": {"name": "BO-Fam:MA-B",
               "current": 981.8604620640567,
               "strength": 2.9171485448187342},
        "TS": {"name": "TS-Fam:MA-B",
               "current": 680.1,
               "strength": 2.9377924330512246},
        "SI": {"name": "SI-Fam:MA-B1B2",
               "current": 394.1,
               "strength": 2.9340661454870647},
        }

    def setUp(self):
        """Create strength object."""
        super().setUp()
        self.str_obj = {}
        for dipole, config in self.settings.items():
            self.str_obj[dipole] = DipoleNormalizer(config["name"])

    def _convert_current(self, dipole):
        strength = self.str_obj[dipole].conv_current_2_strength(
            self.settings[dipole]["current"])
        self.assertAlmostEqual(
            strength, self.settings[dipole]["strength"], places=7)

    def _convert_strength(self, dipole):
        current = self.str_obj[dipole].conv_strength_2_current(
            self.settings[dipole]["strength"])
        self.assertAlmostEqual(
            current, self.settings[dipole]["current"], places=7)

    def test_tb_current_conversion(self):
        """Test TB conversion of current to energy."""
        self._convert_current("TB")

    def test_tb_strength_conversion(self):
        """Test TB conversion of energy to current."""
        self._convert_strength("TB")

    def test_bo_current_conversion(self):
        """Test BO conversion of current to energy."""
        self._convert_current("BO")

    def test_bo_strength_conversion(self):
        """Test BO conversion of energy to current."""
        self._convert_strength("BO")

    def test_ts_current_conversion(self):
        """Test TS conversion of current to energy."""
        self._convert_current("TS")

    def test_ts_strength_conversion(self):
        """Test TS conversion of energy to current."""
        self._convert_strength("TS")

    def test_si_current_conversion(self):
        """Test SI conversion of current to energy."""
        self._convert_current("SI")

    def test_si_strength_conversion(self):
        """Test SI conversion of energy to current."""
        self._convert_strength("SI")



"""Beamline Control."""

from .device import DeviceNC as _DeviceNC


class BLPPSCtrl(_DeviceNC):
    """Beamline Control."""

    class DEVICES:
        """Devices names."""

        CAX = 'CAX'
        ALL = (CAX, )

    _properties = (
        'F:PPS01:FOEOPEN',  # Enviar 1 para abrir o Front-end completo (abre o gamma e o photon)
        'F:PPS01:FOECLOSE',  # Enviar 1 para fechar o Front-end completo (fecha o photon e o gamma)
        'F:PPS01:GS_X_STATUS',  # Status gamma shutter front-end – 0 indica aberto
        'F:PPS01:PS_STATUS',  # Status photon shutter front-end – 0 indica aberto
        'A:PPS01:OEAOPENCLOSE',  # shutter da cabana a - Enviar 1 –se estiver fechado abre – se estiver aberto fecha
        'A:PPS01:PG_STATUS',  # Status do gamma shutter da cabana A – 0 indica aberto
        'M:PPS01:HABILITACAO_MAQUINA',  # Status da liberação do gamma pela máquina – 1 indica liberado
        'A:PPS01:SEARCH_OK',  # status da procura cabana A – 1 indica procura feita
        'B:PPS01:SEARCH_OK',  # status da procura cabana B – 1 indica procura feita
    )

    def __init__(self, devname, *args, **kwargs):
        """Init."""
        # check if device exists
        if devname not in BLPPSCtrl.DEVICES.ALL:
            raise NotImplementedError(devname)
        # call base class constructor
        super().__init__(devname, properties=self._properties, *args, **kwargs)

    @property
    def is_hutchA_intlk_search_done(self):
        """."""
        return self['A:PPS01:SEARCH_OK'] == 1

    @property
    def is_hutchB_intlk_search_done(self):
        """."""
        return self['B:PPS01:SEARCH_OK'] == 1

    @property
    def is_machine_gamma_enabled(self):
        """."""
        return self['M:PPS01:HABILITACAO_MAQUINA'] == 1

    @property
    def is_frontend_gamma_shutter_opened(self):
        """."""
        return self['F:PPS01:GS_X_STATUS'] == 0

    @property
    def is_frontend_photon_shutter_opened(self):
        """."""
        return self['F:PPS01:PS_STATUS'] == 0
    
    @property
    def is_hutchA_gamma_shutter_opened(self):
        """."""
        return self['A:PPS01:PG_STATUS'] == 0
    
    def cmd_frontend_gamma_and_photon_open(self):
        """"."""
        self['F:PPS01:FOEOPEN'] = 1

    def cmd_frontend_gamma_and_photon_close(self):
        """"."""
        self['F:PPS01:FOECLOSE'] = 1

    def cmd_hutchA_gamma_open(self):
        """"."""
        if not self.is_hutchA_gamma_shutter_opened:
            self['A:PPS01:OEAOPENCLOSE'] = 1

    def cmd_hutchA_gamma_close(self):
        """"."""
        if self.is_hutchA_gamma_shutter_opened:
            self['A:PPS01:OEAOPENCLOSE'] = 1

    def cmd_frontend_gamma_and_photon_close(self):
        """"."""
        self['F:PPS01:FOECLOSE'] = 1



"""Define properties of all timing devices and their connections."""

from copy import deepcopy as _dcopy
import numpy as _np
from mathphys import constants as _c

from ..util import ClassProperty as _classproperty
from .. import csdev as _csdev
from ..optics import constants as _oc
from ..search import HLTimeSearch as _HLTimeSearch


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    EVT_MODES = ('Disable', 'Continuous', 'Injection', 'OneShot', 'External')
    TRIG_SRC_LL = (
        'Dsbl', 'Trigger', 'Clock0', 'Clock1', 'Clock2',
        'Clock3', 'Clock4', 'Clock5', 'Clock6', 'Clock7')
    LOCKLL = ('Unlocked', 'Locked')
    DLYTYP = ('Manual', 'Auto')
    DIRECTION = ('Receive', 'Transmit')
    ININJTAB = ('No', 'Yes')
    RFOUT = ('OFF', '5RF/2', '5RF/4', 'RF', 'RF/2', 'RF/4')


_et = ETypes  # syntactic sugar


class Const(_csdev.Const):
    """Constants important for the timing system."""

    AC_FREQUENCY = 60  # [Hz]
    RF_DIVISION = 4
    RF_FREQUENCY = \
        _c.light_speed / _oc.SI.circumference * _oc.SI.harmonic_number
    BASE_FREQUENCY = RF_FREQUENCY / RF_DIVISION
    RF_PERIOD = 1/RF_FREQUENCY
    BASE_DELAY = 1 / BASE_FREQUENCY
    RF_DELAY = BASE_DELAY / 20
    FINE_DELAY = 5e-12  # [s] (five picoseconds)

    EvtModes = _csdev.Const.register('EvtModes', _et.EVT_MODES)
    EvtDlyTyp = _csdev.Const.register('EvtDlyTyp', _et.FIXED_INCR)
    ClockStates = _csdev.Const.register('ClockStates', _et.DSBL_ENBL)
    TrigStates = _csdev.Const.register('TrigStates', _et.DSBL_ENBL)
    TrigPol = _csdev.Const.register('TrigPol', _et.NORM_INV)
    LowLvlLock = _csdev.Const.register('LowLvlLock', _et.LOCKLL)
    TrigDlyTyp = _csdev.Const.register('TrigDlyTyp', _et.DLYTYP)
    TrigDir = _csdev.Const.register('TrigDir', _et.DIRECTION)
    InInjTab = _csdev.Const.register('InInjTab', _et.ININJTAB)
    TrigSrcLL = _csdev.Const.register('TrigSrcLL', _et.TRIG_SRC_LL)
    HLTrigStatusLabels = (
        'All PVs connected',
        'Device Enabled',
        'Fout Enabled',
        'EVG Enabled',
        'Network Ok',
        'UPLink Ok',
        'DownLink Ok',
        'Fout DownLink Ok',
        'EVG DownLink Ok',
        'Interlock Status',
        )

    __EvtHL2LLMap = None
    __EvtLL2HLMap = None

    evt_ll_codes = list(range(64)) + [117, 124]
    evt_ll_names = ['Evt{0:02d}'.format(i) for i in evt_ll_codes]
    EvtLL = _csdev.Const.register(
        'EventsLL', evt_ll_names, values=evt_ll_codes)
    del evt_ll_codes, evt_ll_names  # cleanup class namespace

    ClkHL2LLMap = {
        'Clock0': 'Clk0', 'Clock1': 'Clk1',
        'Clock2': 'Clk2', 'Clock3': 'Clk3',
        'Clock4': 'Clk4', 'Clock5': 'Clk5',
        'Clock6': 'Clk6', 'Clock7': 'Clk7'}
    ClkLL2HLMap = {val: key for key, val in ClkHL2LLMap.items()}

    clk_ll_codes = list(range(8))
    clk_ll_names = ['Clk{0:d}'.format(i) for i in clk_ll_codes]
    ClkLL = _csdev.Const.register(
                    'ClocksLL', clk_ll_names, values=clk_ll_codes)
    del clk_ll_names, clk_ll_codes

    @_classproperty
    def EvtHL2LLMap(cls):
        """."""
        if cls.__EvtHL2LLMap is None:
            cls.__EvtHL2LLMap = _HLTimeSearch.get_hl_events()
            cls.__EvtLL2HLMap = {
                val: key for key, val in cls.__EvtHL2LLMap.items()}
        return cls.__EvtHL2LLMap

    @_classproperty
    def EvtLL2HLMap(cls):
        """."""
        cls.EvtHL2LLMap
        return cls.__EvtLL2HLMap


def get_otp_database(otp_num=0, prefix=None):
    """Return otp_database."""
    def_prefix = 'OTP{0:02d}'.format(otp_num)
    prefix = def_prefix if prefix is None else prefix
    dbase = dict()

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    dbase[prefix+'State-Sts'] = dic_
    dbase[prefix+'State-Sel'] = _dcopy(dic_)

    dic_ = {
        'type': 'int', 'value': 1, 'unit': '',
        'lolo': 1, 'low': 1, 'lolim': 1,
        'hilim': 63, 'high': 63, 'hihi': 63}
    dbase[prefix+'Evt-SP'] = dic_
    dbase[prefix+'Evt-RB'] = _dcopy(dic_)

    dic_ = {
        'type': 'int', 'value': 1, 'unit': '',
        'lolo': 1, 'low': 1, 'lolim': 1,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}
    dbase[prefix+'Width-SP'] = dic_
    dbase[prefix+'Width-RB'] = _dcopy(dic_)

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.NORM_INV}
    dbase[prefix+'Polarity-Sts'] = dic_
    dbase[prefix+'Polarity-Sel'] = _dcopy(dic_)

    dic_ = {
        'type': 'int', 'value': 1, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}
    dbase[prefix+'NrPulses-SP'] = dic_
    dbase[prefix+'NrPulses-RB'] = _dcopy(dic_)

    dic_ = {
        'type': 'int', 'value': 1, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}
    dbase[prefix+'Delay-SP'] = dic_
    dbase[prefix+'Delay-RB'] = _dcopy(dic_)

    return dbase


def get_out_database(out_num=0, equip='EVR', prefix=None):
    """Return out_database."""
    def_prefix = 'OUT{0:d}'.format(out_num)
    prefix = def_prefix if prefix is None else prefix
    dbase = dict()

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.TRIG_SRC_LL}
    dbase[prefix+'Src-Sts'] = dic_
    dbase[prefix+'Src-Sel'] = _dcopy(dic_)

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DLYTYP}
    dbase[prefix+'RFDelayType-Sts'] = dic_
    dbase[prefix+'RFDelayType-Sel'] = _dcopy(dic_)

    max_trig = 23 if equip == 'EVR' else 15
    num_trig = out_num + 12 if equip == 'EVR' else out_num
    dic_ = {
        'type': 'int', 'value': num_trig, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': max_trig, 'high': max_trig, 'hihi': max_trig}
    dbase[prefix+'SrcTrig-SP'] = dic_
    dbase[prefix+'SrcTrig-RB'] = _dcopy(dic_)

    dic_ = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 30, 'high': 30, 'hihi': 30}
    dbase[prefix+'RFDelay-SP'] = dic_
    dbase[prefix+'RFDelay-RB'] = _dcopy(dic_)

    dic_ = {
        'type': 'int', 'value': 1, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 200, 'high': 200, 'hihi': 200}
    dbase[prefix+'FineDelay-SP'] = dic_
    dbase[prefix+'FineDelay-RB'] = _dcopy(dic_)

    return dbase


def get_afc_out_database(out_num=0, out_tp='FMC', prefix=None):
    """Return afc_database."""
    def_prefix = (out_tp + '{0:d}'.format(out_num))
    if out_tp == 'FMC':
        fmc = (out_num // 5) + 1
        ch = (out_num % 5) + 1
        def_prefix = (out_tp + '{0:d}CH{1:d}'.format(fmc, ch))

    prefix = def_prefix if prefix is None else prefix
    dbase = get_otp_database(prefix=prefix)
    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.TRIG_SRC_LL}
    dbase[prefix+'Src-Sts'] = dic_
    dbase[prefix+'Src-Sel'] = _dcopy(dic_)

    return dbase


def get_evr_database(prefix=None):
    """Return evr_database."""
    prefix = prefix or ''
    dbase = dict()

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    dbase[prefix+'DevEnbl-Sts'] = dic_
    dbase[prefix+'DevEnbl-Sel'] = _dcopy(dic_)

    dbase[prefix+'Los-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 255, 'high': 255, 'hihi': 255}

    dbase[prefix+'Alive-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}

    dbase[prefix+'Network-Mon'] = {
        'type': 'enum', 'value': 1,
        'enums': _et.DISCONN_CONN}

    dbase[prefix+'LinkStatus-Mon'] = {
        'type': 'enum', 'value': 1,
        'enums': _et.UNLINK_LINK}

    dbase[prefix+'IntlkStatus-Mon'] = {
        'type': 'enum', 'value': 0,
        'enums': _et.DSBL_ENBL}

    dbase[prefix+'IntlkEnbl-Mon'] = {
        'type': 'enum', 'value': 0,
        'enums': _et.DSBL_ENBL}

    for i in range(24):
        db2 = get_otp_database(otp_num=i)
        for k, v in db2.items():
            dbase[prefix + k] = v

    for i in range(8):
        db2 = get_out_database(out_num=i, equip='EVR')
        for k, v in db2.items():
            dbase[prefix + k] = v
    return dbase


def get_eve_database(eve_num=1, prefix=None):
    """Return eve_database."""
    prefix = prefix or ''
    dbase = dict()

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    dbase[prefix+'DevEnbl-Sts'] = dic_
    dbase[prefix+'DevEnbl-Sel'] = _dcopy(dic_)

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.RFOUT}
    dbase[prefix+'RFOut-Sts'] = dic_
    dbase[prefix+'RFOut-Sel'] = _dcopy(dic_)

    dbase[prefix+'Alive-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}

    dbase[prefix+'Network-Mon'] = {
        'type': 'enum', 'value': 1,
        'enums': _et.DISCONN_CONN}

    dbase[prefix+'LinkStatus-Mon'] = {
        'type': 'enum', 'value': 1,
        'enums': _et.UNLINK_LINK}

    dbase[prefix+'IntlkStatus-Mon'] = {
        'type': 'enum', 'value': 0,
        'enums': _et.DSBL_ENBL}

    dbase[prefix+'IntlkEnbl-Mon'] = {
        'type': 'enum', 'value': 0,
        'enums': _et.DSBL_ENBL}

    for i in range(24):
        db2 = get_otp_database(otp_num=i)
        for k, v in db2.items():
            dbase[prefix + k] = v

    for i in range(8):
        db2 = get_out_database(out_num=i, equip='EVE')
        for k, v in db2.items():
            dbase[prefix + k] = v

    return dbase


def get_afc_database(prefix=None):
    """Return adc_database."""
    prefix = prefix or ''
    dbase = dict()
    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    dbase[prefix+'DevEnbl-Sts'] = dic_
    dbase[prefix+'DevEnbl-Sel'] = _dcopy(dic_)

    dbase[prefix+'Los-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 255, 'high': 255, 'hihi': 255}

    dbase[prefix+'Alive-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}

    dbase[prefix+'Network-Mon'] = {
            'type': 'enum', 'value': 1,
            'enums': _et.DISCONN_CONN}

    dbase[prefix+'LinkStatus-Mon'] = {
            'type': 'enum', 'value': 1,
            'enums': _et.UNLINK_LINK}

    dbase[prefix+'IntlkStatus-Mon'] = {
            'type': 'enum', 'value': 0,
            'enums': _et.DSBL_ENBL}

    for i in range(8):
        db2 = get_afc_out_database(out_num=i, out_tp='AMC')
        for k, v in db2.items():
            dbase[prefix + k] = v

    for i in range(10):
        db2 = get_afc_out_database(out_num=i, out_tp='FMC')
        for k, v in db2.items():
            dbase[prefix + k] = v

    return dbase


def get_fout_database(prefix=None):
    """Return fout_database."""
    prefix = prefix or ''
    dbase = dict()

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    dbase[prefix+'DevEnbl-Sts'] = dic_
    dbase[prefix+'DevEnbl-Sel'] = _dcopy(dic_)

    dbase[prefix+'Los-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 255, 'high': 255, 'hihi': 255}

    dbase[prefix+'Alive-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}

    dbase[prefix+'Network-Mon'] = {
        'type': 'enum', 'value': 1, 'enums': _et.DISCONN_CONN}

    dbase[prefix+'Link-Mon'] = {
        'type': 'enum', 'value': 1, 'enums': _et.UNLINK_LINK}

    dbase[prefix+'Intlk-Mon'] = {
        'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    return dbase


def get_event_database(evt_num=0, prefix=None):
    """Return event_database."""
    def_prefix = 'Evt{0:02d}'.format(evt_num)
    prefix = def_prefix if prefix is None else prefix

    dbase = dict()
    dic_ = {
        'type': 'int', 'value': 0, 'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}
    dbase[prefix + 'Delay-SP'] = _dcopy(dic_)
    dbase[prefix + 'Delay-RB'] = dic_
    dic_ = {'type': 'enum', 'enums': _et.EVT_MODES, 'value': 1}
    dbase[prefix + 'Mode-Sel'] = _dcopy(dic_)
    dbase[prefix + 'Mode-Sts'] = dic_
    dic_ = {'type': 'enum', 'enums': _et.FIXED_INCR, 'value': 1}
    dbase[prefix + 'DelayType-Sel'] = _dcopy(dic_)
    dbase[prefix + 'DelayType-Sts'] = dic_
    dic_ = {'type': 'string', 'value': ''}
    dbase[prefix + 'Desc-SP'] = _dcopy(dic_)
    dbase[prefix + 'Desc-RB'] = dic_
    dbase[prefix + 'ExtTrig-Cmd'] = {'type': 'int', 'value': 0}
    return dbase


def get_clock_database(clock_num=0, prefix=None):
    """Return clock_database."""
    def_prefix = 'Clk{0:d}'.format(clock_num)
    prefix = def_prefix if prefix is None else prefix
    dbase = dict()

    dic_ = {
        'type': 'int', 'value': 124948114, 'lolo': 2, 'low': 2, 'lolim': 2,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}
    dbase[prefix + 'MuxDiv-SP'] = _dcopy(dic_)
    dbase[prefix + 'MuxDiv-RB'] = dic_
    dic_ = {'type': 'enum', 'enums': _et.DSBL_ENBL, 'value': 0}
    dbase[prefix + 'MuxEnbl-Sel'] = _dcopy(dic_)
    dbase[prefix + 'MuxEnbl-Sts'] = dic_
    return dbase


def get_evg_database(prefix=None, only_evg=False):
    """Return evg_database."""
    prefix = prefix or ''
    dbase = dict()

    dic_ = {'type': 'enum', 'value': 0, 'enums': _et.DSBL_ENBL}
    dbase[prefix+'DevEnbl-Sts'] = dic_
    dbase[prefix+'DevEnbl-Sel'] = _dcopy(dic_)

    dic_ = {'type': 'enum', 'enums': _et.DSBL_ENBL, 'value': 0}
    dbase[prefix + 'ContinuousEvt-Sel'] = _dcopy(dic_)
    dbase[prefix + 'ContinuousEvt-Sts'] = dic_

    dic_ = {
        'type': 'int', 'count': 864, 'value': 864*[1],
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 864, 'high': 864, 'hihi': 864}
    dbase[prefix + 'BucketList-SP'] = _dcopy(dic_)
    dbase[prefix + 'BucketList-RB'] = dic_
    dbase[prefix + 'BucketListLen-Mon'] = {
        'type': 'int', 'value': 864,
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 864, 'high': 864, 'hihi': 864}

    dic_ = {'type': 'enum', 'enums': _et.DSBL_ENBL, 'value': 0}
    dbase[prefix + 'InjectionEvt-Sel'] = _dcopy(dic_)
    dbase[prefix + 'InjectionEvt-Sts'] = dic_

    dic_ = {
        'type': 'int', 'value': 0, 'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 100, 'high': 100, 'hihi': 100}
    dbase[prefix + 'RepeatBucketList-SP'] = _dcopy(dic_)
    dbase[prefix + 'RepeatBucketList-RB'] = dic_

    dic_ = {
        'type': 'int', 'value': 30, 'lolo': 1, 'low': 1, 'lolim': 1,
        'hilim': 60, 'high': 60, 'hihi': 60}
    dbase[prefix + 'ACDiv-SP'] = _dcopy(dic_)
    dbase[prefix + 'ACDiv-RB'] = dic_

    dic_ = {
        'type': 'int', 'value': 4, 'lolo': 1, 'low': 1, 'lolim': 1,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}
    dbase[prefix + 'RFDiv-SP'] = _dcopy(dic_)
    dbase[prefix + 'RFDiv-RB'] = dic_

    dbase[prefix+'Los-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 255, 'high': 255, 'hihi': 255}

    dbase[prefix+'Alive-Mon'] = {
        'type': 'int', 'value': 0, 'unit': '',
        'lolo': 0, 'low': 0, 'lolim': 0,
        'hilim': 2**31-1, 'high': 2**31-1, 'hihi': 2**31-1}

    dbase[prefix+'Network-Mon'] = {
        'type': 'enum', 'value': 1, 'enums': _et.DISCONN_CONN}

    dbase[prefix+'RFStatus-Mon'] = {
        'type': 'enum', 'value': 1,
        'enums': ('Loss or Out of Range', 'Normal')}

    dbase[prefix+'StateMachine-Mon'] = {
        'type': 'enum', 'value': 4,
        'enums': (
            'Initializing', 'Stopped', 'Continuous', 'Injection',
            'Preparing Continuous', 'Preparing Injection')
        }

    if only_evg:
        return dbase

    for clc in Const.ClkLL2HLMap:
        dbase.update(get_clock_database(prefix=prefix+clc))
    for ev in Const.EvtLL._fields:
        dbase.update(get_event_database(prefix=prefix+ev))
    return dbase


def get_hl_trigger_database(hl_trigger, prefix=''):
    """Return database of the specified hl_trigger."""
    dbase = dict()
    trig_db = _HLTimeSearch.get_hl_trigger_predef_db(hl_trigger)
    ll_trig_names = _HLTimeSearch.get_ll_trigger_names(hl_trigger)

    dic_ = {'type': 'enum', 'enums': _et.DSBL_ENBL}
    dic_.update(trig_db['State'])
    dbase['State-Sts'] = _dcopy(dic_)
    dbase['State-Sel'] = dic_

    dic_ = {'type': 'enum'}
    dic_.update(trig_db['Src'])
    if _HLTimeSearch.has_clock(hl_trigger):
        clocks = tuple(sorted(Const.ClkHL2LLMap))
        dic_['enums'] += clocks
    dic_['enums'] += ('Invalid', )  # for completeness
    dbase['Src-Sts'] = _dcopy(dic_)
    dbase['Src-Sel'] = dic_

    max_dur = 17e6
    max_wid_raw = 2**31 - 1
    dic_ = {
        'type': 'float', 'unit': 'us', 'prec': 3,
        'lolim': 0.008, 'hilim': max_dur}
    dic_.update(trig_db['Duration'])
    dbase['Duration-RB'] = _dcopy(dic_)
    dbase['Duration-SP'] = dic_

    # Have to be float for spinbox to work properly
    dic_ = {
        'type': 'float', 'unit': 'hard', 'prec': 0, 'value': 0,
        'lolim': 1.0, 'hilim': max_wid_raw}
    dic_.update(trig_db.get('WidthRaw', dict()))
    dbase['WidthRaw-RB'] = _dcopy(dic_)
    dbase['WidthRaw-SP'] = dic_

    dic_ = {'type': 'enum', 'enums': _et.NORM_INV}
    dic_.update(trig_db['Polarity'])
    dbase['Polarity-Sts'] = _dcopy(dic_)
    dbase['Polarity-Sel'] = dic_

    dic_ = {
        'type': 'int', 'unit': 'pulses', 'lolim': 1, 'hilim': 100000}
    dic_.update(trig_db['NrPulses'])
    dbase['NrPulses-RB'] = _dcopy(dic_)
    dbase['NrPulses-SP'] = dic_

    max_dly_raw = 2**31 - 1
    max_dly = 17e6
    dic_ = {
        'type': 'float', 'unit': 'us', 'prec': 3, 'value': 0,
        'lolim': 0.0, 'hilim': max_dly}
    dic_.update(trig_db['Delay'])
    dbase['Delay-RB'] = _dcopy(dic_)
    dbase['Delay-SP'] = dic_

    # Have to be float for spinbox to work properly
    dic_ = {
        'type': 'float', 'unit': 'hard', 'prec': 0, 'value': 0,
        'lolim': 0.0, 'hilim': max_dly_raw}
    dic_.update(trig_db.get('DelayRaw', dict()))
    dbase['DelayRaw-RB'] = _dcopy(dic_)
    dbase['DelayRaw-SP'] = dic_

    dic_ = {
        'type': 'float', 'unit': 'us', 'prec': 3, 'value': 0.0,
        'lolim': 0.0, 'hilim': max_dly}
    dbase['TotalDelay-Mon'] = dic_

    # Have to be float for spinbox to work properly
    dic_ = {
        'type': 'float', 'unit': 'hard', 'prec': 0, 'value': 0,
        'lolim': 0.0, 'hilim': max_dly_raw}
    dbase['TotalDelayRaw-Mon'] = dic_

    siz = len(ll_trig_names)
    dic_ = {
        'type': 'float', 'unit': 'us', 'prec': 3, 'count': siz,
        'value': _np.zeros(siz), 'lolim': -max_dly, 'hilim': max_dly}
    dic_.update(trig_db.get('DeltaDelay', dict()))
    dbase['DeltaDelay-RB'] = _dcopy(dic_)
    dbase['DeltaDelay-SP'] = dic_

    dic_ = {
        'type': 'float', 'unit': 'hard', 'prec': 0,
        'count': siz, 'value': _np.zeros(siz),
        'lolim': -max_dly_raw, 'hilim': max_dly_raw}
    dic_.update(trig_db.get('DeltaDelayRaw', dict()))
    dbase['DeltaDelayRaw-RB'] = _dcopy(dic_)
    dbase['DeltaDelayRaw-SP'] = dic_

    dic_ = {'type': 'enum', 'enums': _et.LOCKLL, 'value': 0}
    dbase['LowLvlLock-Sts'] = _dcopy(dic_)
    dbase['LowLvlLock-Sel'] = dic_

    dic_ = {'type': 'enum', 'enums': _et.DLYTYP}
    dic_.update(trig_db['RFDelayType'])
    if _HLTimeSearch.has_delay_type(hl_trigger):
        dbase['RFDelayType-Sts'] = _dcopy(dic_)
        dbase['RFDelayType-Sel'] = dic_

    dic_ = {'type': 'enum', 'enums': _et.DIRECTION}
    dic_.update(trig_db.get('Direction', dict()))
    if _HLTimeSearch.has_direction(hl_trigger):
        dbase['Direction-Sel'] = _dcopy(dic_)
        dbase['Direction-Sts'] = dic_

    dbase['Status-Mon'] = {'type': 'int', 'value': 0b1111111111}
    dbase['InInjTable-Mon'] = {
        'type': 'enum', 'enums': _et.ININJTAB, 'value': 0}

    dbase['StatusLabels-Cte'] = {
        'type': 'char', 'count': 1000,
        'value': '\n'.join(Const.HLTrigStatusLabels)
        }
    ll_trigs = '\n'.join(ll_trig_names)
    dbase['LowLvlTriggers-Cte'] = {
        'type': 'char', 'count': 5000, 'value': ll_trigs}
    channels = '\n'.join(_HLTimeSearch.get_hl_trigger_channels(hl_trigger))
    dbase['CtrldChannels-Cte'] = {
        'type': 'char', 'count': 5000, 'value': channels}

    return {prefix + pv: dt for pv, dt in dbase.items()}



from .hl_time_search import HLTimeSearch
from .ll_time_search import LLTimeSearch
from .ps_search import PSSearch
from .ma_search import MASearch
from .bpms_search import BPMSearch
from .id_search import IDSearch
from .ioc_search import IOCSearch

del hl_time_search
del ll_time_search
del ps_search
del ma_search
del bpms_search
del id_search
del ioc_search



"""Define PVs, constants and properties of LIEmitMeas IOC."""

from .. import csdev as _csdev


# --- Enumeration Types ---
class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    MEASURESTATE = ('Stopped', 'Measuring')
    PLANE = ('Y', 'X')


_et = ETypes  # syntactic sugar


# --- Const class ---
class Const(_csdev.Const):
    """Const class defining constants."""

    DEFAULT_PROFILE = 'LA-BI:PRF5'

    MeasureState = _csdev.Const.register('MeasureState', _et.MEASURESTATE)
    Plane = _csdev.Const.register('Plane', _et.PLANE)

    @classmethod
    def get_database(cls, prefix=''):
        """Return IOC database."""
        dbase = {
            'MeasureCtrl-Sel': {
                'type': 'enum', 'value': cls.MeasureState.Stopped,
                'enums': cls.MeasureState._fields},
            'MeasureCtrl-Sts': {
                'type': 'enum', 'value': cls.MeasureState.Stopped,
                'enums': cls.MeasureState._fields},
            }
        for val in dbase.values():
            low = val.get('lolim', None)
            hig = val.get('hilim', None)
            if low is not None:
                val['lolo'] = low
                val['low'] = low
            if hig is not None:
                val['hihi'] = hig
                val['high'] = hig
        if prefix:
            dbase = {prefix + k: v for k, v in dbase.items()}
        dbase.update(_csdev.Const.get_database(prefix=prefix))
        return dbase



"""Configurations connection classes."""

import time as _time
import datetime as _datetime
import re as _re
from copy import deepcopy as _dcopy

from .configdb_client import ConfigDBClient as _ConfigDBClient


class ConfigDBDocument():
    """Abstract configuration class."""

    def __init__(self, config_type, name=None, url=None):
        """Constructor."""
        self._configdbclient = \
            _ConfigDBClient(url=url, config_type=config_type)
        self._name = name or self.generate_config_name()
        self._info = None
        self._value = None
        self._synchronized = False

    @property
    def configdbclient(self):
        """."""
        return self._configdbclient

    @property
    def url(self):
        """."""
        return self._configdbclient.url

    @property
    def config_type(self):
        """."""
        return self._configdbclient.config_type

    @property
    def connected(self):
        """."""
        return self._configdbclient.connected

    @property
    def name(self):
        """Name of configuration."""
        return self._name

    @name.setter
    def name(self, value):
        """Set name of configuration."""
        if self._configdbclient.check_valid_configname(value):
            self._name = value
            self._synchronized = False

    @property
    def info(self):
        """Metadata of configuration."""
        return dict(self._info)

    @property
    def created(self):
        """."""
        return self._info['created']

    @property
    def modified(self):
        """."""
        return self._info['modified'][-1]

    @property
    def modified_count(self):
        """."""
        return len(self._info['modified'])

    @property
    def discarded(self):
        """."""
        return self._info['discarded']

    @property
    def value(self):
        """Get configuration."""
        return _dcopy(self._value)

    @value.setter
    def value(self, value):
        """Set configuration."""
        self._set_value(value)

    @property
    def synchronized(self):
        """Return sync state of object and configuration in server."""
        return self._synchronized

    def exist(self):
        """Return True if configuration exists in ConfigServer."""
        info = self._configdbclient.find_configs(name=self._name)
        return len(info) > 0

    def load(self, discarded=False):
        """Load configuration from ConfigServer."""
        self._info = self._configdbclient.get_config_info(
            name=self._name, discarded=discarded)
        self._value = self._configdbclient.get_config_value(
            name=self._name, discarded=discarded)
        self._synchronized = True

    def save(self, new_name=None):
        """Save configuration to ConfigServer."""
        # if config is syncronized, it is not necessary to save an identical
        # one in server
        if self.exist() and self._synchronized and not new_name:
            return

        # check if data format is ok
        if not self._configdbclient.check_valid_value(self._value):
            raise ValueError(
                'Configuration value with inconsistent format.')

        # if new_name is given, apply
        if new_name is not None:
            self._name = new_name

        self._configdbclient.insert_config(self._name, self._value)
        self._info = self._configdbclient.get_config_info(name=self._name)
        self._synchronized = True

    def delete(self):
        """Delete configuration from server."""
        # NOTE: should this method be easily available?
        if self.exist():
            self._configdbclient.delete_config(self._name)
            self._synchronized = False

    def check_valid_value(self, value):
        """."""
        return self._configdbclient.check_valid_value(value)

    def get_value_from_template(self):
        """."""
        return self._configdbclient.get_value_from_template()

    @classmethod
    def generate_config_name(cls, name=None):
        """Generate a configuration name using current imestamp."""
        if name is None:
            name = ''
        name = name.strip()
        # tsf = _re.compile('^\d\d\d\d\d\d-\d\d\d\d\d\d')
        tsf = _re.compile('^[\d]{6}-[\d]{6}')
        if tsf.match(name):
            new_name = cls._get_timestamp() + name[13:]
        else:
            if name:
                new_name = cls._get_timestamp() + ' ' + name
            else:
                new_name = cls._get_timestamp()
        return new_name

    def _set_value(self, value):
        if self.check_valid_value(value):
            self._value = _dcopy(value)
            self._synchronized = False
        else:
            raise ValueError('Invalid value.')

    def _get_item(self, index):
        return

    def _set_item(self, index, value):
        self._synchronized = False
        return

    def __len__(self):
        """Length of configuration."""
        return len(self._value) if hasattr(self._value, '__len__') else 0

    def __getitem__(self, index):
        """Return configuration item."""
        return self._get_item(index)

    def __setitem__(self, index, value):
        """Set configuration item."""
        self._set_item(index, value)

    @staticmethod
    def _get_timestamp(now=None):
        """."""
        if now is None:
            now = _time.time()
        return _datetime.datetime.fromtimestamp(now).strftime('%y%m%d-%H%M%S')



#!/usr/bin/env python-sirius

"""Module to test AS-AP-PosAng Soft IOC main module."""

import unittest
from unittest import mock
import siriuspy.util as util
from siriuspy.posang.main import App


PUB_INTERFACE = (
    'init_database',
    'pvs_database',
    'process',
    'write',
)


class TestASAPPosAngMain(unittest.TestCase):
    """Test AS-AP-PosAng Soft IOC."""

    def setUp(self):
        """Initialize Soft IOC."""
        self.q_ok = {
            'respm-x': [[4.3444644271865913, 0.28861438350278495],
                        [0.93275939885026393, 1.0003702976563984]],
            'respm-y': [[5.6604558827773026, 2.4137865916000418],
                        [1.065850170973988, 0.9767489447759754]]}
        cs_patcher = mock.patch(
            "siriuspy.posang.main._ConfigDBClient", autospec=True)
        self.addCleanup(cs_patcher.stop)
        self.mock_cs = cs_patcher.start()
        self.mock_cs.return_value.get_config_value.return_value = self.q_ok
        pv_patcher = mock.patch("siriuspy.posang.main._PV", autospec=True)
        self.addCleanup(pv_patcher.stop)
        self.mock_pv = pv_patcher.start()
        cnh_patcher = mock.patch(
            "siriuspy.posang.main._HandleConfigNameFile", autospec=True)
        self.addCleanup(cnh_patcher.stop)
        self.mock_cnh = cnh_patcher.start()
        self.mock_cnh.return_value.get_config_name.return_value = \
            'Default_CHSept'
        self.app = App('TB', 'ch-sept')

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            App, PUB_INTERFACE, print_flag=True)
        self.assertTrue(valid)

    def test_write_statuserror_Delta(self):
        """Test write DeltaPosY-SP & DeltaAngY-SP on status error."""
        self.mock_pv.return_value.connected = False
        self.app.write('DeltaPosX-SP', 0.01)
        self.app.write('DeltaAngX-SP', 0.01)
        self.app.write('DeltaPosY-SP', 0.01)
        self.app.write('DeltaAngY-SP', 0.01)
        self.mock_pv.return_value.put.assert_not_called()

    def test_write_ok_SetNewRefKick(self):
        """Test write SetNewRefKick-Cmd in normal operation."""
        self.app.write('SetNewRefKick-Cmd', 0)
        self.assertEqual(self.app._orbx_deltapos, 0)
        self.assertEqual(self.app._orbx_deltaang, 0)
        self.assertEqual(self.app._orby_deltapos, 0)
        self.assertEqual(self.app._orby_deltaang, 0)

    def test_write_ok_ConfigPS(self):
        """Test write ConfigPS-Cmd in normal operation."""
        self.mock_pv.return_value.connected = True
        self.app.write('ConfigPS-Cmd', 0)
        count = self.mock_pv.return_value.put.call_count
        self.assertTrue(count, 2*2)

    def test_write_connerror_Cmds(self):
        """Test write SetNewRefKick-Cmd/ConfigPS-Cmd on connection error."""
        self.mock_pv.return_value.connected = False
        self.app.write('SetNewRefKick-Cmd', 0)
        self.app.write('ConfigPS-Cmd', 0)
        self.mock_pv.return_value.get.assert_not_called()
        self.mock_pv.return_value.put.assert_not_called()


if __name__ == "__main__":
    unittest.main()



"""Load and process BBB to PS data from static table in remote server."""

from copy import deepcopy as _dcopy
from threading import Lock as _Lock
from siriuspy.namesys import Filter as _Filter, SiriusPVName as _PVName
import siriuspy.clientweb as _web

_timeout = 1.0


class BPMSearch:
    """Class with mapping BPMS.

    Data are read from the Sirius web server.
    """

    _mapping = None

    _lock = _Lock()

    # BPMsDATA API
    # ==========
    @classmethod
    def reset(cls):
        """Reload data from files."""
        cls._mapping = None
        cls._get_data()

    @classmethod
    def server_online(cls):
        """Return True/False if Sirius web server is online."""
        return _web.server_online()

    @classmethod
    def is_valid_devname(cls, devname):
        """Check if devname is a valid BPM or PBPM name."""
        cls._get_data()
        return devname in cls._mapping

    @classmethod
    def is_photon_bpm(cls, devname):
        """Check whether BPM is a Phothon BPM (XBPM)."""
        devname = _PVName(devname)
        return devname.dev == 'PBPM'

    @classmethod
    def is_rf_bpm(cls, devname):
        """Check whether BPM is a RF BPM."""
        devname = _PVName(devname)
        return devname.dev == 'BPM'

    @classmethod
    def is_id_rf_bpm(cls, devname):
        """Check whether BPM is from Insertion Device."""
        devname = _PVName(devname)
        return devname.sub.endswith('SA', 'SB', 'SP')

    @classmethod
    def get_mapping(cls):
        """Return a dictionary with the BPMs."""
        cls._get_data()
        return _dcopy(cls._mapping)

    @classmethod
    def get_names(cls, filters=None, sorting=None):
        """Return a list with the bpm names for the given filter."""
        cls._get_data()
        return _Filter.process_filters(
            cls._names, filters=filters, sorting=sorting)

    @classmethod
    def get_nicknames(cls, names=None, filters=None, sorting=None):
        """Return a list with BPM nicknames."""
        cls._get_data()
        if not names:
            names = cls.get_names(filters=filters, sorting=sorting)
        nicknames = len(names)*['']
        for i, bpm in enumerate(names):
            nicknames[i] = bpm.sub + ('-' + bpm.idx if bpm.idx else '')
        return nicknames

    @classmethod
    def get_positions(cls, names=None, filters=None, sorting=None):
        """Return a list with the positions along the ring."""
        cls._get_data()
        if not names:
            names = cls.get_names(filters=filters, sorting=sorting)
        return [cls._mapping[k]['position'] for k in names]

    @classmethod
    def _get_data(cls):
        with cls._lock:
            if cls._mapping is not None:
                return
            if not _web.server_online():
                raise Exception('could not read data from web server!')
            text = _web.bpms_data(timeout=_timeout)
            cls._build_mapping(text)

    @classmethod
    def _build_mapping(cls, text):
        mapping = dict()
        lines = text.splitlines()
        for line in lines:
            line = line.strip()
            if not line or line[0] == '#':
                continue  # empty line
            key, pos, *_ = line.split()
            key = _PVName(key)
            if key in mapping.keys():
                raise Exception('BPM {0:s} double entry.'.format(key))
            else:
                mapping[key] = {'position': float(pos)}
        cls._mapping = mapping
        cls._names = sorted(cls._mapping.keys())



#!/usr/bin/env python-sirius

"""Test reconstrction factories."""

from copy import deepcopy as _dcopy
import argparse as _argparse
import numpy as _np
import matplotlib.pyplot as plt

from siriuspy.clientconfigdb import ConfigDBDocument
from siriuspy.ramp.ramp import BoosterRamp
from siriuspy.ramp.waveform import Waveform
from siriuspy.ramp.magnet import get_magnet
from siriuspy.ramp.reconst_factory import BONormListFactory, BORFRampFactory, \
    BOTIRampFactory, BODipRampFactory

from siriuspy.search import PSSearch


PARSER = _argparse.ArgumentParser(description="Test BORamp Rebuild Factories.")
PARSER.add_argument(
    "factory", type=str, default='', help="Define which factory to test.")
ARGS = PARSER.parse_args()


PSNAMES = PSSearch.get_psnames({'sec': 'BO', 'dis': 'PS'})

R_ORIG = BoosterRamp('testing')
R_ORIG.load()

GLOBAL_CONF = ConfigDBDocument(
    config_type='global_config', name='ref_config')
GLOBAL_CONF.load()


def run():
    if ARGS.factory == 'BONormListFactory':
        dip_wav = R_ORIG.ps_waveform_get('BO-Fam:PS-B-1')

        ps2wfm = dict()
        # for psn in PSNAMES:
        #     if psn in ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2'):
        #         continue
        #     ps2wfm[psn] = R_ORIG.ps_waveform_get_currents(psn)

        for psn in PSNAMES:
            if psn in ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2'):
                continue
            for row in GLOBAL_CONF.value['pvs']:
                pvname = psn + ':Wfm-SP'
                if row[0] == pvname:
                    ps2wfm[psn] = row[1]
                    break

        fac = BONormListFactory(
            ramp_config=R_ORIG, waveforms=ps2wfm,
            opt_global=False, opt_times=False,
            opt_metric='current', use_config_times=True,
            use_straigth_estim=True,
            consider_beam_interval=False)
        # fac.read_waveforms()

        r_built = BoosterRamp('config_test')
        attrs = [
            'ps_ramp_wfm_nrpoints_fams',
            'ps_ramp_wfm_nrpoints_corrs',
            'ps_ramp_duration',
            'ps_ramp_start_energy',
            'ps_ramp_rampup1_start_time',
            'ps_ramp_rampup1_start_energy',
            'ps_ramp_rampup2_start_time',
            'ps_ramp_rampup2_start_energy',
            'ps_ramp_rampup_smooth_intvl',
            'ps_ramp_rampup_smooth_energy',
            'ps_ramp_rampdown_start_time',
            'ps_ramp_rampdown_start_energy',
            'ps_ramp_rampdown_stop_time',
            'ps_ramp_rampdown_stop_energy',
            'ps_ramp_rampdown_smooth_intvl',
            'ps_ramp_rampdown_smooth_energy',
            'rf_ramp_rampup_start_time',
            'rf_ramp_rampup_stop_time',
            'rf_ramp_rampdown_start_time',
            'rf_ramp_rampdown_stop_time',
            'rf_ramp_bottom_voltage',
            'rf_ramp_top_voltage',
            'rf_ramp_bottom_phase',
            'rf_ramp_top_phase',
            'ti_params_injection_time',
            'ti_params_ejection_time',
            'ti_params_ps_ramp_delay',
            'ti_params_rf_ramp_delay',
        ]
        for attr in attrs:
            setattr(r_built, attr, getattr(R_ORIG, attr))

        r_built.ps_normalized_configs_set(fac.normalized_configs)
        ntimes = r_built.ps_normalized_configs_times
        print(r_built)

        des_prec = fac.desired_reconstr_precision
        prec_reached = fac.precision_reached
        print('Precision rechead?', prec_reached[0])
        print('Maximum abs.error:', prec_reached[1])

        max_error = 0.0

        axs = plt.gca()
        axs.grid()
        for psn in PSNAMES:
            if psn in ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2'):
                continue
            strgs = list()
            for tim in ntimes:
                strgs.append(r_built[tim][psn])
            strgs = _np.array(strgs)

            nr_pts = R_ORIG.ps_ramp_wfm_nrpoints_fams if 'Fam' in psn \
                else R_ORIG.ps_ramp_wfm_nrpoints_corrs
            orig_times = R_ORIG.ps_waveform_get_times(psn)
            orig_currs = ps2wfm[psn]
            orig_strgs = Waveform(
                psn, dipole=dip_wav, currents=orig_currs,
                wfm_nrpoints=nr_pts).strengths
            built_times = r_built.ps_waveform_get_times(psn)
            built_currs = r_built.ps_waveform_get_currents(psn)
            built_strgs = r_built.ps_waveform_get_strengths(psn)
            dif_curr = orig_currs - built_currs
            dif_strg = orig_strgs - built_strgs
            max_error = max(max_error, max(dif_curr))
            if not _np.allclose(dif_curr, 0, atol=des_prec):
                # axs.plot(orig_times, orig_strgs, 'r.-', label=psn)
                # axs.plot(built_times, built_strgs, 'k.-')
                # axs.plot(ntimes, strgs, 'b.')
                # axs.plot(built_times, dif_curr, label=psn)
                axs.plot(built_times, dif_strg, label=psn)

        # for tim in ntimes:
        #     axs.plot([tim, tim], [-max_error, max_error], 'k-.')
        plt.legend()
        plt.title('Curves with error > than desired ({})'.format(des_prec))
        plt.show()

    elif ARGS.factory == 'BORFRampFactory':
        fac = BORFRampFactory()
        params = fac.rf_params
        _print_param_dict(params)

    elif ARGS.factory == 'BOTIRampFactory':
        fac = BOTIRampFactory()
        params = fac.ti_params
        _print_param_dict(params)

    elif ARGS.factory == 'BODipRampFactory':
        dip_name = 'BO-Fam:PS-B-1'
        dip_mag = get_magnet('BO-Fam:MA-B')

        # wav_orig = R_ORIG.ps_waveform_get_currents(dip_name)

        for row in GLOBAL_CONF.value['pvs']:
            if row[0] == dip_name + ':Wfm-SP':
                wav_orig = row[1]
                break

        wav_energy = dip_mag.conv_current_2_strength(wav_orig)

        fac = BODipRampFactory(ramp_config=R_ORIG, waveform=wav_orig)
        # fac.read_waveforms()
        params_final = fac.dip_params
        _print_param_dict(params_final)

        w_built = _dcopy(R_ORIG.ps_waveform_get(dip_name))
        for name, value in params_final.items():
            setattr(w_built, name, value)
        wav_curr = w_built.currents
        err_curr = wav_orig - wav_curr
        wav_strgt = w_built.strengths
        err_strgt = wav_energy - wav_strgt

        success = _np.allclose(err_curr, 0, atol=1e-5)
        max_error = _np.max(_np.abs(err_curr))
        print('Precision rechead?', success)
        print('Maximum abs.error:', max_error)

        fig, axs = plt.subplots(1, 2)
        fig.set_size_inches(12, 6)
        axs[0].plot(err_curr)
        axs[0].grid()
        axs[0].set_title('Current error')
        axs[1].plot(err_strgt)
        axs[1].grid()
        axs[1].set_title('Strength error')
        plt.show()


def _print_param_dict(params):
    string = ''
    fmtstr = '{:<'+str(25)+'}: {}\n'
    for key in sorted(params):
        string += fmtstr.format(key, params[key])
    print(string)


if __name__ == "__main__":
    run()



"""Simulated PV."""


from ..namesys import SiriusPVName as _SiriusPVName
from ..epics.pv_fake import add_to_database as _add_to_database
from ..epics.pv_fake import PVFake as _PVFake

from . import DBASE_DEFAULT as _DBASE_DEF
from .simulation import Simulation as _Simulation


class SimPV(_PVFake):
    """."""

    DBASE_DEFAULT = _DBASE_DEF

    def __new__(cls, pvname, *args, **kwargs):
        """Return existing SimPV object or return a new one."""
        _, _ = args, kwargs  # throwaway arguments
        if not _Simulation.register_state_get():
            raise Exception('Simulation registration of SimPVs is Off!')
        instance = _Simulation.pv_find(pvname)
        if not instance:
            instance = super(SimPV, cls).__new__(cls)
        return instance

    def __init__(
            self, pvname, **kwargs):
        """."""
        # if object already exists, no initialization needed.
        if _Simulation.pv_find(pvname):
            return

        # --- initializations ---

        # convert string to SiriusPVName
        pvname = _SiriusPVName(pvname)

        # get pv database
        db_pv = _Simulation.pv_dbase_find(pvname, unique=True)
        dbase = {pvname: db_pv if db_pv else SimPV.DBASE_DEFAULT}

        # init pv database
        _add_to_database(dbase, '')

        # call base class constructor
        super().__init__(pvname, **kwargs)

        # register SimPV in simulation
        _Simulation.pv_register(self)

    @property
    def value(self):
        """."""
        _ = _Simulation.pv_get(self.pvname)
        return super().get()

    @value.setter
    def value(self, value):
        """."""
        status = _Simulation.pv_put(self.pvname, value)
        if status:
            super().put(value)

    def get(self, **kwargs):
        """."""
        _ = _Simulation.pv_get(self.pvname, **kwargs)
        return super().get(**kwargs)

    def put(self, value, **kwargs):
        """."""
        status = _Simulation.pv_put(self.pvname, value, **kwargs)
        if status:
            super().put(value, **kwargs)

    def get_sim(self):
        """Get SimPV value without invoking simulator callback."""
        return super().get()

    def put_sim(self, value):
        """Set SimPV value without invoking simulator callback."""
        super().put(value)



"""Beam stability info database."""

import os as _os

from .. import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    STAB_UNSTAB = ('Stable', 'Unstable')


_et = ETypes


# --- Const class ---

class StabInfoConst(_csdev.Const):
    """Stability Info Const class."""

    StabUnstab = _csdev.Const.register('StabUnstab', _et.STAB_UNSTAB)

    SI_HARMNUM = 864

    CURR_THRES = 5  # [mA]

    DEF_BBBH_CALIBFACTOR = 1.718  # [counts/mA/um]
    DEF_BBBV_CALIBFACTOR = 4.917  # [counts/mA/um]
    DEF_BBBL_CALIBFACTOR = 1000  # [counts/mA/degree]

    BBBH_OSCAMP_THRES = 5  # [um]
    BBBV_OSCAMP_THRES = 1.5  # [um]
    BBBL_OSCAMP_THRES = 0.5  # [degree]

    def __init__(self):
        """Init."""
        ioc_fol = _os.path.join(
            '/home', 'sirius', 'iocs-log', 'si-ap-stabinfo', 'data')
        self.autosave_fname = _os.path.join(ioc_fol, 'autosave.txt')

    def get_propty_database(self):
        """Return property database."""
        dbase = {
            'Version-Cte': {'type': 'str', 'value': 'UNDEF'},
            'Log-Mon': {'type': 'string', 'value': 'Starting...'},

            'BbBHCalibFactor-SP': {
                'type': 'float', 'value': self.DEF_BBBH_CALIBFACTOR, 'prec': 3,
                'unit': 'counts/mA/um', 'lolim': -10.0, 'hilim': 10.0},
            'BbBHCalibFactor-RB': {
                'type': 'float', 'value': self.DEF_BBBH_CALIBFACTOR, 'prec': 3,
                'unit': 'counts/mA/um', 'lolim': -10.0, 'hilim': 10.0},
            'BbBHOscAmpThres-Cte': {
                'type': 'float', 'value': self.BBBH_OSCAMP_THRES, 'prec': 3,
                'unit': 'um'},
            'BbBHOscAmp-Mon': {
                'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'um'},
            'BbBHStatus-Mon': {
                'type': 'enum', 'value': self.StabUnstab.Stable,
                'enums': _et.STAB_UNSTAB},

            'BbBVCalibFactor-SP': {
                'type': 'float', 'value': self.DEF_BBBV_CALIBFACTOR, 'prec': 3,
                'unit': 'counts/mA/um', 'lolim': -10.0, 'hilim': 10.0},
            'BbBVCalibFactor-RB': {
                'type': 'float', 'value': self.DEF_BBBV_CALIBFACTOR, 'prec': 3,
                'unit': 'counts/mA/um', 'lolim': -10.0, 'hilim': 10.0},
            'BbBVOscAmpThres-Cte': {
                'type': 'float', 'value': self.BBBV_OSCAMP_THRES, 'prec': 3,
                'unit': 'um'},
            'BbBVOscAmp-Mon': {
                'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'um'},
            'BbBVStatus-Mon': {
                'type': 'enum', 'value': self.StabUnstab.Stable,
                'enums': _et.STAB_UNSTAB},

            'BbBLCalibFactor-SP': {
                'type': 'float', 'value': self.DEF_BBBL_CALIBFACTOR, 'prec': 3,
                'unit': 'counts/mA/degree', 'lolim': 0.0, 'hilim': 10000.0},
            'BbBLCalibFactor-RB': {
                'type': 'float', 'value': self.DEF_BBBL_CALIBFACTOR, 'prec': 3,
                'unit': 'counts/mA/degree', 'lolim': 0.0, 'hilim': 10000.0},
            'BbBLOscAmpThres-Cte': {
                'type': 'float', 'value': self.BBBL_OSCAMP_THRES, 'prec': 3,
                'unit': 'degree'},
            'BbBLOscAmp-Mon': {
                'type': 'float', 'value': 0.0, 'prec': 3, 'unit': 'degree'},
            'BbBLStatus-Mon': {
                'type': 'enum', 'value': self.StabUnstab.Stable,
                'enums': _et.STAB_UNSTAB},
        }
        dbase = _csdev.add_pvslist_cte(dbase)
        return dbase



"""Insertion Device Feedforward Configuration."""

from copy import deepcopy as _dcopy

import numpy as _np

from ..search import IDSearch as _IDSearch
from ..clientconfigdb import ConfigDBDocument as _ConfigDBDocument


class IDFFConfig(_ConfigDBDocument):
    """Insertion Device Feedforward Configuration."""

    # NOTE: for EPU50 there is a large discrepancy
    # between RB/SP/Mon phase values
    PPARAM_TOL = 0.5  # [mm]
    KPARAM_TOL = 0.1  # [mm]
    CONFIGDB_TYPE = 'si_idff'

    def __init__(self, name=None, url=None):
        """."""
        name_ = name or 'idff_' + self.generate_config_name()
        self._polarization_definitions = None
        super().__init__(
            config_type=IDFFConfig.CONFIGDB_TYPE, name=name_, url=url)

    @property
    def pparameter_pvname(self):
        """Return ID pparameter pvname."""
        config = self._value
        if config:
            kparm = config['pvnames']['pparameter']
            return kparm
        else:
            raise ValueError('Configuration not defined!')

    @property
    def kparameter_pvname(self):
        """Return ID kparameter pvname."""
        config = self._value
        if config:
            kparm = config['pvnames']['kparameter']
            return kparm
        else:
            raise ValueError('Configuration not defined!')

    @property
    def ch_pvnames(self):
        """Return CH corrector power supply pvnames."""
        return self._get_corr_pvnames('ch1', 'ch2')

    @property
    def cv_pvnames(self):
        """Return CV corrector power supply pvnames."""
        return self._get_corr_pvnames('cv1', 'cv2')

    @property
    def qs_pvnames(self):
        """Return QS corrector power supply pvnames."""
        return self._get_corr_pvnames('qs1', 'qs2')

    @property
    def polarizations(self):
        """Return list of light polarizations in the IDFF config."""
        if self._value:
            return list(self._value['polarizations'].keys())
        else:
            raise ValueError('Configuration not defined!')

    @property
    def value(self):
        """Get configuration."""
        return _dcopy(self._value)

    @value.setter
    def value(self, value):
        """Set configuration."""
        self._set_value(value)

    def calculate_setpoints(self, polarization, kparameter_value):
        """Return correctors setpoints for a particular ID config.

        The parameter 'kparameter' can be a gap or phase value,
        depending on the insertion device.
        """
        if self._value:
            setpoints = dict()
            idff = self._value['polarizations'][polarization]
            kparameter_values = idff['kparameter']
            setpoints = dict()
            for corrlabel, table in idff.items():
                if corrlabel not in ('pparameter', 'kparameter'):
                    # linear interpolation
                    setpoint = _np.interp(
                        kparameter_value, kparameter_values, table)
                    corr_pvname = self._value['pvnames'][corrlabel]
                    setpoints[corr_pvname] = setpoint
            return setpoints
        else:
            raise ValueError('Configuration not defined!')

    @staticmethod
    def create_template_config(idname):
        """Create a template configuration for a given ID type."""
        idff = _IDSearch.conv_idname_2_idff(idname)

        # description
        description = 'Config created with IDFFConfig.create_template_config'

        # build value for pvnames key
        pvnames = dict()
        for key, corr_pvname in idff.items():
            if key != 'polarizations':
                pvnames[key] = corr_pvname

        # build value for polarizations key
        ctable = dict()
        for pvname in pvnames:
            ctable[pvname] = [0, 0]
        ptable = dict(ctable)
        ptable['pparameter'] = [0, 0]
        ptable['kparameter'] = 0
        ktable = dict(ctable)
        ktable['pparameter'] = 0
        ktable['kparameter'] = [0, 0]

        polarizations = dict()
        for polarization in idff['polarizations']:
            if polarization == 'none':
                polarizations[polarization] = dict(ptable)
            else:
                polarizations[polarization] = dict(ktable)

        template_config = dict(
            description=description, pvnames=pvnames,
            polarizations=polarizations)
        return template_config

    def __str__(self):
        """."""
        stg = ''
        stg += f'name: {self.name}'
        value = self.value
        if value is None:
            return stg

        stg += '\n--- pvnames ---'
        pvnames = ''.join(
            [f'\n{key}: {value}' for key, value in value['pvnames'].items()])
        stg += pvnames

        for pol, table in self.value['polarizations'].items():
            stg += f'\n--- {pol} ---'
            for key, value in table.items():
                if isinstance(value, (int, float)):
                    nrpts = 1
                    str_ = f'{value}'
                else:
                    nrpts = len(value)
                    str_ = f'{min(value)}..{max(value)} [{nrpts}]'
                stg += f'\n{key} : ' + str_
        return stg

    def load(self, discarded=False):
        """."""
        super().load(discarded=discarded)
        self._calc_polariz_defs()

    def get_polarization_state(self, pparameter, kparameter):
        """Return polarization state based on ID parameteres."""
        poldefs = self._polarization_definitions
        if poldefs is None:
            raise ValueError('No IDFF configuration defined.')
        for pol, val in poldefs.items():
            if pol == 'none':
                continue
            if val is None or abs(pparameter - val) < IDFFConfig.PPARAM_TOL:
                return pol
        if abs(kparameter - poldefs['none']) < IDFFConfig.KPARAM_TOL:
            return 'none'
        return 'not_defined'

    def check_valid_value(self, value):
        """."""
        if not super().check_valid_value(value):
            return False
        for pol, table in value['polarizations'].items():
            if pol == 'none':
                nrpts = len(table['pparameter'])
            else:
                nrpts = len(table['kparameter'])
            for key, value_ in table.items():
                if key in ('kparameter', 'pparameter'):
                    continue
                if len(value_) != nrpts:
                    return False
        return True

    def _get_corr_pvnames(self, cname1, cname2):
        """Return corrector power supply pvnames."""
        if self._value:
            pvnames = self._value['pvnames']
            corr1, corr2 = pvnames.get(cname1), pvnames.get(cname2)
            return corr1, corr2
        else:
            raise ValueError('Configuration not defined!')

    def _set_value(self, value):
        super()._set_value(value)
        self._calc_polariz_defs()

    def _calc_polariz_defs(self):
        """."""
        data = self._value['polarizations']
        poldefs = dict()
        for pol, tab in data.items():
            if pol != 'none':
                poldefs[pol] = tab['pparameter']
            else:
                poldefs[pol] = tab['kparameter']
        self._polarization_definitions = poldefs
